"""Deck configuration resource provider."""

from typing import List, Set, Tuple

from opentrons_shared_data.deck.types import (
    DeckDefinitionV5,
    CutoutFixture,
)

from opentrons.types import DeckSlotName

from ..types import (
    AddressableArea,
    AreaType,
    PotentialCutoutFixture,
    DeckPoint,
    Dimensions,
    AddressableOffsetVector,
)
from ..errors import (
    CutoutDoesNotExistError,
    FixtureDoesNotExistError,
    AddressableAreaDoesNotExistError,
    SlotDoesNotExistError,
)


def get_cutout_position(cutout_id: str, deck_definition: DeckDefinitionV5) -> DeckPoint:
    """Get the base position of a cutout on the deck."""
    for cutout in deck_definition["locations"]["cutouts"]:
        if cutout_id == cutout["id"]:
            position = cutout["position"]
            return DeckPoint(x=position[0], y=position[1], z=position[2])
    else:
        raise CutoutDoesNotExistError(f"Could not find cutout with name {cutout_id}")


def get_cutout_fixture(
    cutout_fixture_id: str, deck_definition: DeckDefinitionV5
) -> CutoutFixture:
    """Gets cutout fixture from deck that matches the cutout fixture ID provided."""
    for cutout_fixture in deck_definition["cutoutFixtures"]:
        if cutout_fixture["id"] == cutout_fixture_id:
            return cutout_fixture
    raise FixtureDoesNotExistError(
        f"Could not find cutout fixture with name {cutout_fixture_id}"
    )


def get_provided_addressable_area_names(
    cutout_fixture_id: str, cutout_id: str, deck_definition: DeckDefinitionV5
) -> List[str]:
    """Gets a list of the addressable areas provided by the cutout fixture on the cutout."""
    cutout_fixture = get_cutout_fixture(cutout_fixture_id, deck_definition)
    try:
        return cutout_fixture["providesAddressableAreas"][cutout_id]
    except KeyError:
        return []


def get_addressable_area_display_name(
    addressable_area_name: str, deck_definition: DeckDefinitionV5
) -> str:
    """Get the display name for an addressable area name."""
    for addressable_area in deck_definition["locations"]["addressableAreas"]:
        if addressable_area["id"] == addressable_area_name:
            return addressable_area["displayName"]
    raise AddressableAreaDoesNotExistError(
        f"Could not find addressable area with name {addressable_area_name}"
    )


def get_potential_cutout_fixtures(
    addressable_area_name: str, deck_definition: DeckDefinitionV5
) -> Tuple[str, Set[PotentialCutoutFixture]]:
    """Given an addressable area name, gets the cutout ID associated with it and a set of potential fixtures."""
    potential_fixtures = []
    for cutout_fixture in deck_definition["cutoutFixtures"]:
        for cutout_id, provided_areas in cutout_fixture[
            "providesAddressableAreas"
        ].items():
            if addressable_area_name in provided_areas:
                potential_fixtures.append(
                    PotentialCutoutFixture(
                        cutout_id=cutout_id,
                        cutout_fixture_id=cutout_fixture["id"],
                        provided_addressable_areas=frozenset(provided_areas),
                    )
                )
    # This following logic is making the assumption that every addressable area can only go on one cutout, though
    # it may have multiple cutout fixtures that supply it on that cutout. If this assumption changes, some of the
    # following logic will have to be readjusted
    if not potential_fixtures:
        raise AddressableAreaDoesNotExistError(
            f"{addressable_area_name} is not provided by any cutout fixtures"
            f" in deck definition {deck_definition['otId']}"
        )
    cutout_id = potential_fixtures[0].cutout_id
    assert all(cutout_id == fixture.cutout_id for fixture in potential_fixtures)
    return cutout_id, set(potential_fixtures)


def get_addressable_area_from_name(
    addressable_area_name: str,
    cutout_position: DeckPoint,
    deck_definition: DeckDefinitionV5,
) -> AddressableArea:
    """Given a name and a cutout position, get an addressable area on the deck."""
    for addressable_area in deck_definition["locations"]["addressableAreas"]:
        if addressable_area["id"] == addressable_area_name:
            cutout_id, _ = get_potential_cutout_fixtures(
                addressable_area_name, deck_definition
            )
            base_slot = get_deck_slot_for_cutout_id(cutout_id)
            area_offset = addressable_area["offsetFromCutoutFixture"]
            position = AddressableOffsetVector(
                x=area_offset[0] + cutout_position.x,
                y=area_offset[1] + cutout_position.y,
                z=area_offset[2] + cutout_position.z,
            )
            bounding_box = Dimensions(
                x=addressable_area["boundingBox"]["xDimension"],
                y=addressable_area["boundingBox"]["yDimension"],
                z=addressable_area["boundingBox"]["zDimension"],
            )

            return AddressableArea(
                area_name=addressable_area["id"],
                area_type=AreaType(addressable_area["areaType"]),
                base_slot=base_slot,
                display_name=addressable_area["displayName"],
                bounding_box=bounding_box,
                position=position,
                compatible_module_types=addressable_area.get(
                    "compatibleModuleTypes", []
                ),
            )
    raise AddressableAreaDoesNotExistError(
        f"Could not find addressable area with name {addressable_area_name}"
    )


def get_deck_slot_for_cutout_id(cutout_id: str) -> DeckSlotName:
    """Get the corresponding deck slot for an addressable area."""
    try:
        return CUTOUT_TO_DECK_SLOT_MAP[cutout_id]
    except KeyError as e:
        raise CutoutDoesNotExistError(
            f"Could not find data for cutout {cutout_id}"
        ) from e


def get_cutout_id_by_deck_slot_name(slot_name: DeckSlotName) -> str:
    """Get the Cutout ID of a given Deck Slot by Deck Slot Name."""
    try:
        return DECK_SLOT_TO_CUTOUT_MAP[slot_name]
    except KeyError as e:
        raise SlotDoesNotExistError(
            f"Could not find data for slot {slot_name.value}"
        ) from e


def get_labware_hosting_addressable_area_name_for_cutout_and_cutout_fixture(
    cutout_id: str, cutout_fixture_id: str, deck_definition: DeckDefinitionV5
) -> str:
    """Get the first addressable area that can contain labware for a cutout and fixture.

    This probably isn't relevant outside of labware offset locations, where (for now) nothing
    provides more than one labware-containing addressable area.
    """
    for cutoutFixture in deck_definition["cutoutFixtures"]:
        if cutoutFixture["id"] != cutout_fixture_id:
            continue
        provided_aas = cutoutFixture["providesAddressableAreas"].get(cutout_id, None)
        if provided_aas is None:
            raise CutoutDoesNotExistError(
                f"{cutout_fixture_id} does not go in {cutout_id}"
            )
        for aa_id in provided_aas:
            for addressable_area in deck_definition["locations"]["addressableAreas"]:
                if addressable_area["id"] != aa_id:
                    continue
                # TODO: In deck def v6 this will be easier, but as of right now there isn't really
                # a way to tell from an addressable area whether it takes labware so let's take the
                # first one
                return aa_id
            raise AddressableAreaDoesNotExistError(
                f"Could not find an addressable area that allows labware from cutout fixture {cutout_fixture_id} in cutout {cutout_id}"
            )

    raise FixtureDoesNotExistError(f"Could not find entry for {cutout_fixture_id}")


# This is a temporary shim while Protocol Engine's conflict-checking code
# can only take deck slots as input.
# Long-term solution: Check for conflicts based on bounding boxes, not slot adjacencies.
# Shorter-term: Change the conflict-checking code to take cutouts instead of deck slots.
CUTOUT_TO_DECK_SLOT_MAP: dict[str, DeckSlotName] = {
    # OT-2
    "cutout1": DeckSlotName.SLOT_1,
    "cutout2": DeckSlotName.SLOT_2,
    "cutout3": DeckSlotName.SLOT_3,
    "cutout4": DeckSlotName.SLOT_4,
    "cutout5": DeckSlotName.SLOT_5,
    "cutout6": DeckSlotName.SLOT_6,
    "cutout7": DeckSlotName.SLOT_7,
    "cutout8": DeckSlotName.SLOT_8,
    "cutout9": DeckSlotName.SLOT_9,
    "cutout10": DeckSlotName.SLOT_10,
    "cutout11": DeckSlotName.SLOT_11,
    "cutout12": DeckSlotName.FIXED_TRASH,
    # Flex
    "cutoutA1": DeckSlotName.SLOT_A1,
    "cutoutA2": DeckSlotName.SLOT_A2,
    "cutoutA3": DeckSlotName.SLOT_A3,
    "cutoutB1": DeckSlotName.SLOT_B1,
    "cutoutB2": DeckSlotName.SLOT_B2,
    "cutoutB3": DeckSlotName.SLOT_B3,
    "cutoutC1": DeckSlotName.SLOT_C1,
    "cutoutC2": DeckSlotName.SLOT_C2,
    "cutoutC3": DeckSlotName.SLOT_C3,
    "cutoutD1": DeckSlotName.SLOT_D1,
    "cutoutD2": DeckSlotName.SLOT_D2,
    "cutoutD3": DeckSlotName.SLOT_D3,
}
DECK_SLOT_TO_CUTOUT_MAP = {
    deck_slot: cutout for cutout, deck_slot in CUTOUT_TO_DECK_SLOT_MAP.items()
}



import asyncio
import os
import re
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from typing import Any, Optional, List, Dict, Tuple

from .hid_protocol import (
    AbsorbanceHidInterface as AbsProtocol,
    ErrorCodeNames,
    DeviceStateNames,
    SlotStateNames,
    MeasurementConfig,
)
from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderPlatePresence,
    AbsorbanceReaderDeviceState,
    ABSMeasurementMode,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control.modules.errors import AbsorbanceReaderDisconnectedError


SN_PARSER = re.compile(r'ATTRS{serial}=="(?P<serial>.+?)"')
# match semver V0.0.0 (old format) or one integer (latest format)
VERSION_PARSER = re.compile(r"(?P<version>(V\d+\.\d+\.\d+|^\d+$))")
SERIAL_PARSER = re.compile(r"(?P<serial>(OPT|BYO)[A-Z]{3}[0-9]+)")


class AsyncByonoy:
    """Async wrapper around Byonoy Device Library."""

    @staticmethod
    def match_device_with_sn(
        sn: str, devices: List[AbsProtocol.Device]
    ) -> AbsProtocol.Device:
        for device in devices:
            if device.sn == sn:
                return device
        raise RuntimeError(f"Unavailble module with serial number: {sn}")

    @staticmethod
    def serial_number_from_port(name: str) -> str:
        """
        Get the serial number from a port using pyusb.
        """
        import usb.core as usb_core  # type: ignore[import-untyped]

        port_numbers = tuple(int(s) for s in name.split("-")[1].split("."))
        device = usb_core.find(port_numbers=port_numbers)
        if device:
            return str(device.serial_number)
        raise RuntimeError(f"Could not find serial number for port: {name}")

    @classmethod
    async def create(
        cls,
        port: str,
        usb_port: USBPort,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> "AsyncByonoy":
        """
        Create an AsyncByonoy instance.

        Args:
            port: url or port name
            baud_rate: the baud rate
            timeout: optional timeout in seconds
            write_timeout: optional write timeout in seconds
            loop: optional event loop. if None get_running_loop will be used
            reset_buffer_before_write: reset the serial input buffer before
             writing to it
        """
        loop = loop or asyncio.get_running_loop()
        executor = ThreadPoolExecutor(max_workers=1)

        import byonoy_devices as byonoy  # type: ignore[import-not-found]

        interface: AbsProtocol = byonoy

        device_sn = cls.serial_number_from_port(usb_port.name)
        found: List[AbsProtocol.Device] = await loop.run_in_executor(
            executor=executor, func=byonoy.available_devices
        )
        device = cls.match_device_with_sn(device_sn, found)

        return cls(
            interface=interface,
            device=device,
            executor=executor,
            loop=loop,
        )

    def __init__(
        self,
        interface: AbsProtocol,
        device: AbsProtocol.Device,
        executor: ThreadPoolExecutor,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """
        Constructor

        Args:
            serial: connected Serial object
            executor: a thread pool executor
            loop: event loop
        """
        self._interface = interface
        self._device = device
        self._executor = executor
        self._loop = loop
        self._supported_wavelengths: Optional[list[int]] = None
        self._device_handle: Optional[int] = None
        self._current_config: Optional[MeasurementConfig] = None

    async def open(self) -> bool:
        """
        Open the connection.

        Returns: boolean denoting connection success.
        """

        err, device_handle = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.open_device, self._device),
        )
        self._raise_if_error(err.name, f"Error opening device: {err}")
        self._device_handle = device_handle
        return bool(device_handle)

    async def close(self) -> None:
        """Close the connection."""
        handle = self._verify_device_handle()
        await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.free_device, handle),
        )
        self._device_handle = None

    async def is_open(self) -> bool:
        """True if connection is open."""
        if self._device_handle is None:
            return False
        handle = self._verify_device_handle()
        return await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.device_open, handle),
        )

    async def get_device_information(self) -> Dict[str, str]:
        """Get serial number and version info."""
        handle = self._verify_device_handle()
        err, device_info = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.get_device_information, handle),
        )
        self._raise_if_error(err.name, f"Error getting device information: {err}")
        serial_match = SERIAL_PARSER.match(device_info.sn)
        version_match = VERSION_PARSER.search(device_info.version)
        serial = serial_match["serial"].strip() if serial_match else "OPTMAA00000"
        version = version_match["version"].lower() if version_match else "v0"
        info = {
            "serial": serial,
            "version": version,
            "model": "ABS96",
        }
        return info

    async def get_device_status(self) -> AbsorbanceReaderDeviceState:
        """Get state information of the device."""
        handle = self._verify_device_handle()
        err, status = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.get_device_status, handle),
        )
        self._raise_if_error(err.name, f"Error getting device status: {err}")
        return self.convert_device_state(status.name)

    async def update_firmware(self, firmware_file_path: str) -> Tuple[bool, str]:
        """Updates the firmware of the device."""
        handle = self._verify_device_handle()
        if not os.path.exists(firmware_file_path):
            return False, f"Firmware file not found: {firmware_file_path}"
        err = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.update_device, handle, firmware_file_path),
        )
        if err.name != "NO_ERROR":
            return False, f"Byonoy update failed with error: {err}"
        return True, ""

    async def get_device_uptime(self) -> int:
        """Get how long in seconds the device has been running for."""
        handle = self._verify_device_handle()
        err, uptime = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.get_device_uptime, handle),
        )
        self._raise_if_error(err.name, "Error getting device uptime: ")
        return uptime

    async def get_lid_status(self) -> AbsorbanceReaderLidStatus:
        """Get the state of the absorbance lid."""
        handle = self._verify_device_handle()
        err, lid_info = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.get_device_parts_aligned, handle),
        )
        self._raise_if_error(err.name, f"Error getting lid status: {err}")
        return (
            AbsorbanceReaderLidStatus.ON if lid_info else AbsorbanceReaderLidStatus.OFF
        )

    async def get_supported_wavelengths(self) -> list[int]:
        """Get a list of the wavelength readings this device supports."""
        handle = self._verify_device_handle()
        err, wavelengths = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.abs96_get_available_wavelengths, handle),
        )
        self._raise_if_error(err.name, "Error getting available wavelengths: ")
        self._supported_wavelengths = wavelengths
        return wavelengths

    async def get_measurement(self) -> List[List[float]]:
        """Gets one or more measurements based on the current configuration."""
        handle = self._verify_device_handle()
        assert (
            self._current_config is not None
        ), "Cannot get measurement without initializing."
        measure_func: Any = self._interface.abs96_single_measure
        if isinstance(self._current_config, AbsProtocol.MultiMeasurementConfig):
            measure_func = self._interface.abs96_multiple_measure
        err, measurements = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(
                measure_func,
                handle,
                self._current_config,
            ),
        )
        self._raise_if_error(err.name, f"Error getting measurement: {err}")
        return measurements if isinstance(measurements[0], List) else [measurements]  # type: ignore

    async def get_plate_presence(self) -> AbsorbanceReaderPlatePresence:
        """Get the state of the plate for the reader."""
        handle = self._verify_device_handle()
        err, presence = await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._interface.get_device_slot_status, handle),
        )
        self._raise_if_error(err.name, f"Error getting slot status: {err}")
        return self.convert_plate_presence(presence.name)

    def _get_supported_wavelengths(self) -> List[int]:
        handle = self._verify_device_handle()
        wavelengths: List[int]
        err, wavelengths = self._interface.abs96_get_available_wavelengths(handle)
        self._raise_if_error(err.name, f"Error getting available wavelengths: {err}")
        self._supported_wavelengths = wavelengths
        return wavelengths

    def _initialize_measurement(self, conf: MeasurementConfig) -> None:
        handle = self._verify_device_handle()
        if isinstance(conf, AbsProtocol.SingleMeasurementConfig):
            err = self._interface.abs96_initialize_single_measurement(handle, conf)
        else:
            err = self._interface.abs96_initialize_multiple_measurement(handle, conf)
        self._raise_if_error(err.name, f"Error initializing measurement: {err}")
        self._current_config = conf

    def _initialize(
        self,
        mode: ABSMeasurementMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        if not self._supported_wavelengths:
            self._get_supported_wavelengths()
        assert self._supported_wavelengths
        conf: MeasurementConfig
        if set(wavelengths).issubset(self._supported_wavelengths):
            if mode == ABSMeasurementMode.SINGLE:
                conf = self._interface.Abs96SingleMeasurementConfig()
                conf.sample_wavelength = wavelengths[0] or 0
                conf.reference_wavelength = reference_wavelength or 0
            else:
                conf = self._interface.Abs96MultipleMeasurementConfig()
                conf.sample_wavelengths = wavelengths
        else:
            raise ValueError(
                f"Unsupported wavelength: {wavelengths}, expected: {self._supported_wavelengths}"
            )
        self._initialize_measurement(conf)

    async def initialize(
        self,
        mode: ABSMeasurementMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """initialize the device so we can start reading samples from it."""
        await self._loop.run_in_executor(
            executor=self._executor,
            func=partial(self._initialize, mode, wavelengths, reference_wavelength),
        )

    def _verify_device_handle(self) -> int:
        assert self._device_handle is not None, RuntimeError(
            "Device handle not set up."
        )
        return self._device_handle

    def _raise_if_error(
        self,
        err_name: ErrorCodeNames,
        msg: str = "Error occurred: ",
    ) -> None:
        if err_name in [
            "DEVICE_CLOSED",
            "DEVICE_COMMUNICATION_FAILURE",
            "UNSUPPORTED_OPERATION",
        ]:
            raise AbsorbanceReaderDisconnectedError(self._device.sn)
        if err_name != "NO_ERROR":
            raise RuntimeError(msg, err_name)

    @staticmethod
    def convert_device_state(
        device_state: DeviceStateNames,
    ) -> AbsorbanceReaderDeviceState:
        state_map: Dict[DeviceStateNames, AbsorbanceReaderDeviceState] = {
            "UNKNOWN": AbsorbanceReaderDeviceState.UNKNOWN,
            "OK": AbsorbanceReaderDeviceState.OK,
            "BROKEN_FW": AbsorbanceReaderDeviceState.BROKEN_FW,
            "ERROR": AbsorbanceReaderDeviceState.ERROR,
        }
        return state_map[device_state]

    @staticmethod
    def convert_plate_presence(
        slot_state: SlotStateNames,
    ) -> AbsorbanceReaderPlatePresence:
        state_map: Dict[SlotStateNames, AbsorbanceReaderPlatePresence] = {
            "UNKNOWN": AbsorbanceReaderPlatePresence.UNKNOWN,
            "EMPTY": AbsorbanceReaderPlatePresence.ABSENT,
            "OCCUPIED": AbsorbanceReaderPlatePresence.PRESENT,
            "UNDETERMINED": AbsorbanceReaderPlatePresence.UNKNOWN,
        }
        return state_map[slot_state]



"""Structures to represent changes that commands want to make to engine state."""

from __future__ import annotations

import dataclasses
import enum
import typing
from typing_extensions import Self
from datetime import datetime

from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.protocol_engine.resources import pipette_data_provider
from opentrons.protocol_engine.types import (
    DeckPoint,
    LabwareLocation,
    OnLabwareLocation,
    TipGeometry,
    AspiratedFluid,
    LiquidClassRecord,
    ABSMeasureMode,
    LiquidTrackingType,
)
from opentrons.types import MountType
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.pipette.types import PipetteNameType


class _NoChangeEnum(enum.Enum):
    NO_CHANGE = enum.auto()


NO_CHANGE: typing.Final = _NoChangeEnum.NO_CHANGE
"""A sentinel value to indicate that a value shouldn't be changed.

Useful when `None` is semantically unclear or already has some other meaning.
"""


NoChangeType: typing.TypeAlias = typing.Literal[_NoChangeEnum.NO_CHANGE]
"""The type of `NO_CHANGE`, as `NoneType` is to `None`.

Unfortunately, mypy doesn't let us write `Literal[NO_CHANGE]`. Use this instead.
"""


class _ClearEnum(enum.Enum):
    CLEAR = enum.auto()


CLEAR: typing.Final = _ClearEnum.CLEAR
"""A sentinel value to indicate that a value should be cleared.

Useful when `None` is semantically unclear or has some other meaning.
"""


ClearType: typing.TypeAlias = typing.Literal[_ClearEnum.CLEAR]
"""The type of `CLEAR`, as `NoneType` is to `None`.

Unfortunately, mypy doesn't let us write `Literal[CLEAR]`. Use this instead.
"""


@dataclasses.dataclass(frozen=True)
class Well:
    """Designates a well in a labware."""

    labware_id: str
    well_name: str


@dataclasses.dataclass(frozen=True)
class AddressableArea:
    """Designates an addressable area."""

    addressable_area_name: str


@dataclasses.dataclass
class PipetteLocationUpdate:
    """An update to a pipette's location."""

    pipette_id: str
    """The ID of the already-loaded pipette."""

    new_location: Well | AddressableArea | None | NoChangeType
    """The pipette's new logical location.

    Note: `new_location=None` means "change the location to `None` (unknown)",
    not "do not change the location".
    """

    new_deck_point: DeckPoint | NoChangeType


@dataclasses.dataclass
class LabwareLocationUpdate:
    """An update to a labware's location."""

    labware_id: str
    """The ID of the already-loaded labware."""

    new_location: LabwareLocation
    """The labware's new location."""

    offset_id: str | None
    """The ID of the labware's new offset, for its new location."""


@dataclasses.dataclass
class BatchLabwareLocationUpdate:
    """An update to the locations of multiple labware."""

    new_locations_by_id: dict[str, LabwareLocation]
    """The new locations of each ID."""
    new_offset_ids_by_id: dict[str, str | None]
    """The new offsets of each id."""


# todo(mm, 2025-04-28): Combine with BatchLoadedLabwareUpdate.
@dataclasses.dataclass
class LoadedLabwareUpdate:
    """An update that loads a new labware."""

    labware_id: str
    """The unique ID of the new labware."""

    new_location: LabwareLocation
    """The labware's initial location."""

    offset_id: str | None
    """The ID of the labware's offset."""

    display_name: str | None

    definition: LabwareDefinition


@dataclasses.dataclass
class BatchLoadedLabwareUpdate:
    """An update that loads multiple new labware."""

    new_locations_by_id: typing.Dict[str, LabwareLocation]
    """Each new labwares's initial location keyed by Labware ID."""

    offset_ids_by_id: typing.Dict[str, str | None]
    """The ID of each labware's offset keyed by labware ID."""

    display_names_by_id: typing.Dict[str, str | None]
    """The Display Name for each new labware keyed by labware ID"""

    definitions_by_id: typing.Dict[str, LabwareDefinition]
    """The Labware Definition for each labware keyed by Labware ID."""


@dataclasses.dataclass
class LoadedLidStackUpdate:
    """An update that loads a new lid stack."""

    stack_id: str
    """The unique ID of the Lid Stack Object."""

    stack_object_definition: LabwareDefinition
    "The System-only Labware Definition of the Lid Stack Object"

    stack_location: LabwareLocation
    "The initial location of the Lid Stack Object."

    new_locations_by_id: typing.Dict[str, OnLabwareLocation]
    """Each lid's initial location keyed by Labware ID."""

    definition: LabwareDefinition | None
    "The Labware Definition of the Lid Labware(s) loaded."


@dataclasses.dataclass
class LabwareLidUpdate:
    """An update that identifies a lid on a given parent labware."""

    parent_labware_ids: typing.List[str]
    """The unique IDs of the parent labwares."""

    lid_ids: typing.List[str | None]
    """The unique IDs of the new lids."""


@dataclasses.dataclass
class LoadPipetteUpdate:
    """An update that loads a new pipette.

    NOTE: Currently, if this is provided, a PipetteConfigUpdate must always be
    provided alongside it to fully initialize everything.
    """

    pipette_id: str
    """The unique ID of the new pipette."""

    pipette_name: PipetteNameType
    mount: MountType
    liquid_presence_detection: bool | None


@dataclasses.dataclass
class PipetteConfigUpdate:
    """An update to a pipette's config."""

    pipette_id: str
    """The ID of the already-loaded pipette."""

    # todo(mm, 2024-10-14): Does serial_number belong in LoadPipetteUpdate?
    serial_number: str

    config: pipette_data_provider.LoadedStaticPipetteData


@dataclasses.dataclass
class PipetteNozzleMapUpdate:
    """Update pipette nozzle map."""

    pipette_id: str
    nozzle_map: NozzleMap


@dataclasses.dataclass
class PipetteTipStateUpdate:
    """Update pipette tip state."""

    pipette_id: str
    tip_geometry: TipGeometry | None


@dataclasses.dataclass
class PipetteAspirateReadyUpdate:
    """Update pipette ready state."""

    pipette_id: str
    ready_to_aspirate: bool


@dataclasses.dataclass
class TipsUsedUpdate:
    """Represents an update that marks tips in a tip rack as used."""

    labware_id: str
    """The labware ID of the tip rack."""

    well_names: list[str]
    """The exact wells in the tip rack that should be marked as used.

    This is the *full* list, which is probably more than what appeared in the pickUpTip
    command's params, for multi-channel reasons.
    """


@dataclasses.dataclass
class LiquidLoadedUpdate:
    """An update from loading a liquid."""

    labware_id: str
    volumes: typing.Dict[str, float]
    last_loaded: datetime


@dataclasses.dataclass
class LiquidProbedUpdate:
    """An update from probing a liquid."""

    labware_id: str
    well_name: str
    last_probed: datetime
    height: LiquidTrackingType | ClearType
    volume: LiquidTrackingType | ClearType


@dataclasses.dataclass
class LiquidOperatedUpdate:
    """An update from operating a liquid."""

    labware_id: str
    well_names: list[str]
    volume_added: float | ClearType


@dataclasses.dataclass
class PipetteAspiratedFluidUpdate:
    """Represents the pipette aspirating something. Might be air or liquid from a well."""

    pipette_id: str
    fluid: AspiratedFluid
    type: typing.Literal["aspirated"] = "aspirated"


@dataclasses.dataclass
class PipetteEjectedFluidUpdate:
    """Represents the pipette pushing something out. Might be air or liquid."""

    pipette_id: str
    volume: float
    type: typing.Literal["ejected"] = "ejected"


@dataclasses.dataclass
class PipetteUnknownFluidUpdate:
    """Represents the amount of fluid in the pipette becoming unknown."""

    pipette_id: str
    type: typing.Literal["unknown"] = "unknown"


@dataclasses.dataclass
class PipetteEmptyFluidUpdate:
    """Sets the pipette to be valid and empty."""

    pipette_id: str
    clean_tip: bool
    type: typing.Literal["empty"] = "empty"


@dataclasses.dataclass
class AbsorbanceReaderLidUpdate:
    """An update to an absorbance reader's lid location."""

    is_lid_on: bool


@dataclasses.dataclass
class AbsorbanceReaderDataUpdate:
    """An update to an absorbance reader's lid location."""

    read_result: typing.Dict[int, typing.Dict[str, float]]


@dataclasses.dataclass(frozen=True)
class AbsorbanceReaderInitializeUpdate:
    """An update to an absorbance reader's initialization."""

    measure_mode: ABSMeasureMode
    sample_wave_lengths: typing.List[int]
    reference_wave_length: typing.Optional[int]


@dataclasses.dataclass
class AbsorbanceReaderStateUpdate:
    """An update to the absorbance reader module state."""

    module_id: str
    absorbance_reader_lid: AbsorbanceReaderLidUpdate | NoChangeType = NO_CHANGE
    absorbance_reader_data: AbsorbanceReaderDataUpdate | NoChangeType = NO_CHANGE
    initialize_absorbance_reader_update: (
        AbsorbanceReaderInitializeUpdate | NoChangeType
    ) = NO_CHANGE


@dataclasses.dataclass
class FlexStackerPoolConstraint:
    """The labware definitions that are contained in the pool."""

    max_pool_count: int
    primary_definition: LabwareDefinition
    lid_definition: LabwareDefinition | None
    adapter_definition: LabwareDefinition | None


@dataclasses.dataclass
class FlexStackerStateUpdate:
    """An update to the Flex Stacker module state."""

    module_id: str
    pool_constraint: FlexStackerPoolConstraint | NoChangeType = NO_CHANGE
    pool_count: int | NoChangeType = NO_CHANGE

    @classmethod
    def create_or_override(
        cls,
        maybe_inst: FlexStackerStateUpdate | NoChangeType,
        module_id: str,
    ) -> FlexStackerStateUpdate:
        """Build or default a state update."""
        if maybe_inst == NO_CHANGE:
            return FlexStackerStateUpdate(module_id=module_id)
        else:
            return maybe_inst


@dataclasses.dataclass
class LiquidClassLoadedUpdate:
    """The state update from loading a liquid class."""

    liquid_class_id: str
    liquid_class_record: LiquidClassRecord


@dataclasses.dataclass
class FilesAddedUpdate:
    """An update that adds a new data file."""

    file_ids: list[str]


@dataclasses.dataclass
class AddressableAreaUsedUpdate:
    """An update that says an addressable area has been used."""

    addressable_area_name: str


@dataclasses.dataclass
class StateUpdate:
    """Represents an update to perform on engine state."""

    pipette_location: PipetteLocationUpdate | NoChangeType | ClearType = NO_CHANGE

    loaded_pipette: LoadPipetteUpdate | NoChangeType = NO_CHANGE

    pipette_config: PipetteConfigUpdate | NoChangeType = NO_CHANGE

    pipette_nozzle_map: PipetteNozzleMapUpdate | NoChangeType = NO_CHANGE

    pipette_tip_state: PipetteTipStateUpdate | NoChangeType = NO_CHANGE

    pipette_aspirated_fluid: (
        PipetteAspiratedFluidUpdate
        | PipetteEjectedFluidUpdate
        | PipetteUnknownFluidUpdate
        | PipetteEmptyFluidUpdate
        | NoChangeType
    ) = NO_CHANGE

    labware_location: LabwareLocationUpdate | NoChangeType = NO_CHANGE

    batch_labware_location: BatchLabwareLocationUpdate | NoChangeType = NO_CHANGE

    loaded_labware: LoadedLabwareUpdate | NoChangeType = NO_CHANGE

    batch_loaded_labware: BatchLoadedLabwareUpdate | NoChangeType = NO_CHANGE

    loaded_lid_stack: LoadedLidStackUpdate | NoChangeType = NO_CHANGE

    labware_lid: LabwareLidUpdate | NoChangeType = NO_CHANGE

    tips_used: TipsUsedUpdate | NoChangeType = NO_CHANGE

    liquid_loaded: LiquidLoadedUpdate | NoChangeType = NO_CHANGE

    liquid_probed: LiquidProbedUpdate | NoChangeType = NO_CHANGE

    liquid_operated: LiquidOperatedUpdate | NoChangeType = NO_CHANGE

    absorbance_reader_state_update: AbsorbanceReaderStateUpdate | NoChangeType = (
        NO_CHANGE
    )

    flex_stacker_state_update: FlexStackerStateUpdate | NoChangeType = NO_CHANGE

    liquid_class_loaded: LiquidClassLoadedUpdate | NoChangeType = NO_CHANGE

    files_added: FilesAddedUpdate | NoChangeType = NO_CHANGE

    addressable_area_used: AddressableAreaUsedUpdate | NoChangeType = NO_CHANGE

    ready_to_aspirate: PipetteAspirateReadyUpdate | NoChangeType = NO_CHANGE

    def append(self, other: Self) -> Self:
        """Apply another `StateUpdate` "on top of" this one.

        This object is mutated in-place, taking values from `other`.
        If an attribute in `other` is `NO_CHANGE`, the value in this object is kept.
        """
        fields = dataclasses.fields(other)
        for field in fields:
            other_value = other.__dict__[field.name]
            if other_value != NO_CHANGE:
                self.__dict__[field.name] = other_value
        return self

    @classmethod
    def reduce(cls: typing.Type[Self], *args: Self) -> Self:
        """Fuse multiple state updates into a single one.

        State updates that are later in the parameter list are preferred to those that are earlier;
        NO_CHANGE is ignored.
        """
        accumulator = cls()
        for arg in args:
            accumulator.append(arg)
        return accumulator

    # These convenience functions let the caller avoid the boilerplate of constructing a
    # complicated dataclass tree, and allow chaining.

    @typing.overload
    def set_pipette_location(
        self: Self, *, pipette_id: str, new_deck_point: DeckPoint
    ) -> Self:
        """Schedule a pipette's coordinates to be changed while preserving its logical location."""

    @typing.overload
    def set_pipette_location(
        self: Self,
        *,
        pipette_id: str,
        new_labware_id: str,
        new_well_name: str,
        new_deck_point: DeckPoint,
    ) -> Self:
        """Schedule a pipette's location to be set to a well."""

    @typing.overload
    def set_pipette_location(
        self: Self,
        *,
        pipette_id: str,
        new_addressable_area_name: str,
        new_deck_point: DeckPoint,
    ) -> Self:
        """Schedule a pipette's location to be set to an addressable area."""
        pass

    def set_pipette_location(  # noqa: D102
        self: Self,
        *,
        pipette_id: str,
        new_labware_id: str | NoChangeType = NO_CHANGE,
        new_well_name: str | NoChangeType = NO_CHANGE,
        new_addressable_area_name: str | NoChangeType = NO_CHANGE,
        new_deck_point: DeckPoint,
    ) -> Self:
        if new_addressable_area_name != NO_CHANGE:
            self.pipette_location = PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=AddressableArea(
                    addressable_area_name=new_addressable_area_name
                ),
                new_deck_point=new_deck_point,
            )
        elif new_labware_id == NO_CHANGE or new_well_name == NO_CHANGE:
            self.pipette_location = PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=NO_CHANGE,
                new_deck_point=new_deck_point,
            )
        else:
            self.pipette_location = PipetteLocationUpdate(
                pipette_id=pipette_id,
                new_location=Well(labware_id=new_labware_id, well_name=new_well_name),
                new_deck_point=new_deck_point,
            )
        return self

    def clear_all_pipette_locations(self) -> Self:
        """Mark all pipettes as having an unknown location."""
        self.pipette_location = CLEAR
        return self

    def set_labware_location(
        self: Self,
        *,
        labware_id: str,
        new_location: LabwareLocation,
        new_offset_id: str | None,
    ) -> Self:
        """Set a labware's location. See `LabwareLocationUpdate`."""
        self.labware_location = LabwareLocationUpdate(
            labware_id=labware_id,
            new_location=new_location,
            offset_id=new_offset_id,
        )
        return self

    def set_batch_labware_location(
        self: Self,
        *,
        new_locations_by_id: typing.Dict[str, LabwareLocation],
        new_offset_ids_by_id: typing.Dict[str, str | None],
    ) -> Self:
        """Update the location of multiple labware objects."""
        self.batch_labware_location = BatchLabwareLocationUpdate(
            new_locations_by_id=new_locations_by_id,
            new_offset_ids_by_id=new_offset_ids_by_id,
        )
        return self

    def set_loaded_labware(
        self: Self,
        definition: LabwareDefinition,
        labware_id: str,
        offset_id: str | None,
        display_name: str | None,
        location: LabwareLocation,
    ) -> Self:
        """Add a new labware to state. See `LoadedLabwareUpdate`."""
        self.loaded_labware = LoadedLabwareUpdate(
            definition=definition,
            labware_id=labware_id,
            offset_id=offset_id,
            new_location=location,
            display_name=display_name,
        )
        return self

    def set_batch_loaded_labware(
        self: Self,
        definitions_by_id: typing.Dict[str, LabwareDefinition],
        offset_ids_by_id: typing.Dict[str, str | None],
        display_names_by_id: typing.Dict[str, str | None],
        new_locations_by_id: typing.Dict[str, LabwareLocation],
    ) -> Self:
        """Add a set of new labwares to state. See `BatchLoadedLabwareUpdate`."""
        self.batch_loaded_labware = BatchLoadedLabwareUpdate(
            new_locations_by_id=new_locations_by_id,
            offset_ids_by_id=offset_ids_by_id,
            display_names_by_id=display_names_by_id,
            definitions_by_id=definitions_by_id,
        )
        return self

    def set_loaded_lid_stack(
        self: Self,
        stack_id: str,
        stack_object_definition: LabwareDefinition,
        stack_location: LabwareLocation,
        locations: typing.Dict[str, OnLabwareLocation],
        labware_definition: LabwareDefinition | None,
    ) -> Self:
        """Add a new lid stack to state. See `LoadedLidStackUpdate`."""
        self.loaded_lid_stack = LoadedLidStackUpdate(
            stack_id=stack_id,
            stack_object_definition=stack_object_definition,
            stack_location=stack_location,
            definition=labware_definition,
            new_locations_by_id=locations,
        )
        return self

    def set_lids(
        self: Self,
        parent_labware_ids: typing.List[str],
        lid_ids: typing.List[str | None],
    ) -> Self:
        """Update the labware parent of a loaded or moved lid. See `LabwareLidUpdate`."""
        self.labware_lid = LabwareLidUpdate(
            parent_labware_ids=parent_labware_ids,
            lid_ids=lid_ids,
        )
        return self

    def set_load_pipette(
        self: Self,
        pipette_id: str,
        pipette_name: PipetteNameType,
        mount: MountType,
        liquid_presence_detection: bool | None,
    ) -> Self:
        """Add a new pipette to state. See `LoadPipetteUpdate`."""
        self.loaded_pipette = LoadPipetteUpdate(
            pipette_id=pipette_id,
            pipette_name=pipette_name,
            mount=mount,
            liquid_presence_detection=liquid_presence_detection,
        )
        return self

    def update_pipette_config(
        self: Self,
        pipette_id: str,
        config: pipette_data_provider.LoadedStaticPipetteData,
        serial_number: str,
    ) -> Self:
        """Update a pipette's config. See `PipetteConfigUpdate`."""
        self.pipette_config = PipetteConfigUpdate(
            pipette_id=pipette_id, config=config, serial_number=serial_number
        )
        return self

    def update_pipette_nozzle(
        self: Self, pipette_id: str, nozzle_map: NozzleMap
    ) -> Self:
        """Update a pipette's nozzle map. See `PipetteNozzleMapUpdate`."""
        self.pipette_nozzle_map = PipetteNozzleMapUpdate(
            pipette_id=pipette_id, nozzle_map=nozzle_map
        )
        return self

    def update_pipette_tip_state(
        self: Self, pipette_id: str, tip_geometry: TipGeometry | None
    ) -> Self:
        """Update a pipette's tip state. See `PipetteTipStateUpdate`."""
        self.pipette_tip_state = PipetteTipStateUpdate(
            pipette_id=pipette_id, tip_geometry=tip_geometry
        )
        return self

    def mark_tips_as_used(self: Self, labware_id: str, well_names: list[str]) -> Self:
        """Mark tips in a tip rack as used. See `TipsUsedUpdate`."""
        self.tips_used = TipsUsedUpdate(labware_id=labware_id, well_names=well_names)
        return self

    def set_liquid_loaded(
        self: Self,
        labware_id: str,
        volumes: typing.Dict[str, float],
        last_loaded: datetime,
    ) -> Self:
        """Add liquid volumes to well state. See `LoadLiquidUpdate`."""
        self.liquid_loaded = LiquidLoadedUpdate(
            labware_id=labware_id,
            volumes=volumes,
            last_loaded=last_loaded,
        )
        return self

    def set_liquid_probed(
        self: Self,
        labware_id: str,
        well_name: str,
        last_probed: datetime,
        height: LiquidTrackingType | ClearType,
        volume: LiquidTrackingType | ClearType,
    ) -> Self:
        """Add a liquid height and volume to well state. See `ProbeLiquidUpdate`."""
        self.liquid_probed = LiquidProbedUpdate(
            labware_id=labware_id,
            well_name=well_name,
            height=height,
            volume=volume,
            last_probed=last_probed,
        )
        return self

    def set_liquid_operated(
        self: Self,
        labware_id: str,
        well_names: list[str],
        volume_added: float | ClearType,
    ) -> Self:
        """Update liquid volumes in well state. See `OperateLiquidUpdate`."""
        self.liquid_operated = LiquidOperatedUpdate(
            labware_id=labware_id,
            well_names=well_names,
            volume_added=volume_added,
        )
        return self

    def set_fluid_aspirated(self: Self, pipette_id: str, fluid: AspiratedFluid) -> Self:
        """Update record of fluid held inside a pipette. See `PipetteAspiratedFluidUpdate`."""
        self.pipette_aspirated_fluid = PipetteAspiratedFluidUpdate(
            type="aspirated", pipette_id=pipette_id, fluid=fluid
        )
        return self

    def set_fluid_ejected(self: Self, pipette_id: str, volume: float) -> Self:
        """Update record of fluid held inside a pipette. See `PipetteEjectedFluidUpdate`."""
        self.pipette_aspirated_fluid = PipetteEjectedFluidUpdate(
            type="ejected", pipette_id=pipette_id, volume=volume
        )
        return self

    def set_fluid_unknown(self: Self, pipette_id: str) -> Self:
        """Update record of fluid held inside a pipette. See `PipetteUnknownFluidUpdate`."""
        self.pipette_aspirated_fluid = PipetteUnknownFluidUpdate(
            type="unknown", pipette_id=pipette_id
        )
        return self

    def set_fluid_empty(self: Self, pipette_id: str, clean_tip: bool = False) -> Self:
        """Update record of fluid held inside a pipette. See `PipetteEmptyFluidUpdate`."""
        self.pipette_aspirated_fluid = PipetteEmptyFluidUpdate(
            type="empty", pipette_id=pipette_id, clean_tip=clean_tip
        )
        return self

    def set_absorbance_reader_lid(self: Self, module_id: str, is_lid_on: bool) -> Self:
        """Update an absorbance reader's lid location. See `AbsorbanceReaderLidUpdate`."""
        assert self.absorbance_reader_state_update == NO_CHANGE
        self.absorbance_reader_state_update = AbsorbanceReaderStateUpdate(
            module_id=module_id,
            absorbance_reader_lid=AbsorbanceReaderLidUpdate(is_lid_on=is_lid_on),
        )
        return self

    def set_absorbance_reader_data(
        self, module_id: str, read_result: typing.Dict[int, typing.Dict[str, float]]
    ) -> Self:
        """Update an absorbance reader's read data. See `AbsorbanceReaderReadDataUpdate`."""
        assert self.absorbance_reader_state_update == NO_CHANGE
        self.absorbance_reader_state_update = AbsorbanceReaderStateUpdate(
            module_id=module_id,
            absorbance_reader_data=AbsorbanceReaderDataUpdate(read_result=read_result),
        )
        return self

    def initialize_absorbance_reader(
        self,
        module_id: str,
        measure_mode: ABSMeasureMode,
        sample_wave_lengths: typing.List[int],
        reference_wave_length: typing.Optional[int],
    ) -> Self:
        """Initialize absorbance reader."""
        assert self.absorbance_reader_state_update == NO_CHANGE
        self.absorbance_reader_state_update = AbsorbanceReaderStateUpdate(
            module_id=module_id,
            initialize_absorbance_reader_update=AbsorbanceReaderInitializeUpdate(
                measure_mode=measure_mode,
                sample_wave_lengths=sample_wave_lengths,
                reference_wave_length=reference_wave_length,
            ),
        )
        return self

    def set_addressable_area_used(self: Self, addressable_area_name: str) -> Self:
        """Mark that an addressable area has been used. See `AddressableAreaUsedUpdate`."""
        self.addressable_area_used = AddressableAreaUsedUpdate(
            addressable_area_name=addressable_area_name
        )
        return self

    def update_flex_stacker_labware_pool_definition(
        self,
        module_id: str,
        max_count: int,
        primary_definition: LabwareDefinition,
        adapter_definition: LabwareDefinition | None,
        lid_definition: LabwareDefinition | None,
    ) -> Self:
        """Constrain the labware pool to a specific definition."""
        self.flex_stacker_state_update = dataclasses.replace(
            FlexStackerStateUpdate.create_or_override(
                self.flex_stacker_state_update, module_id
            ),
            pool_constraint=FlexStackerPoolConstraint(
                max_pool_count=max_count,
                primary_definition=primary_definition,
                lid_definition=lid_definition,
                adapter_definition=adapter_definition,
            ),
        )
        return self

    def update_flex_stacker_labware_pool_count(
        self, module_id: str, count: int
    ) -> Self:
        """Set the labware pool to a specific count."""
        self.flex_stacker_state_update = dataclasses.replace(
            FlexStackerStateUpdate.create_or_override(
                self.flex_stacker_state_update, module_id
            ),
            pool_count=count,
        )
        return self

    def set_pipette_ready_to_aspirate(
        self, pipette_id: str, ready_to_aspirate: bool
    ) -> Self:
        """Set the ready to aspirate state for a pipette."""
        self.ready_to_aspirate = PipetteAspirateReadyUpdate(
            pipette_id=pipette_id, ready_to_aspirate=ready_to_aspirate
        )
        return self



"""Legacy Well core implementation."""
from typing import Optional, Union

from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN

from opentrons.protocols.api_support.util import APIVersionError

from opentrons.types import Point, Mount

from opentrons.protocol_engine.types.liquid_level_detection import (
    SimulatedProbeResult,
    LiquidTrackingType,
)

from .well_geometry import WellGeometry
from ..well import AbstractWellCore
from ..._liquid import Liquid


class LegacyWellCore(AbstractWellCore):
    """Well implementation core based on legacy PAPIv2 behavior.

    Args:
        well_geometry: Information about the well's geometry.
        display_name: User-facing display name.
        has_tip: Whether the well contains a tip.
        name: Unique well name inside its labware, e.g. `A1`.
            Must match pattern /^([A-Z]+)([0-9]+)$/.
    """

    def __init__(
        self, well_geometry: WellGeometry, display_name: str, has_tip: bool, name: str
    ) -> None:
        self._display_name = display_name
        self._has_tip = has_tip
        self._name = name

        match = WELL_NAME_PATTERN.match(name)
        assert match, f"could not match '{name}' using pattern '{WELL_NAME_PATTERN}'"
        self._row_name = match.group(1)
        self._column_name = match.group(2)
        self._geometry = well_geometry

    @property
    def geometry(self) -> WellGeometry:
        """Get the well's geometry information interface."""
        return self._geometry

    @geometry.setter
    def geometry(self, well_geometry: WellGeometry) -> None:
        """Update the well's geometry interface."""
        self._geometry = well_geometry

    @property
    def diameter(self) -> Optional[float]:
        """Get the well's diameter, if circular."""
        return self._geometry.diameter

    @property
    def length(self) -> Optional[float]:
        """Get the well's length, if rectangular."""
        return self._geometry.length

    @property
    def width(self) -> Optional[float]:
        """Get the well's width, if rectangular."""
        return self._geometry.width

    @property
    def depth(self) -> float:
        """Get the well's depth."""
        return self._geometry.depth

    def has_tip(self) -> bool:
        """Whether the well contains a tip."""
        return self._has_tip

    def set_has_tip(self, value: bool) -> None:
        """Set the well as containing or not containing a tip."""
        self._has_tip = value

    def get_display_name(self) -> str:
        """Get the full display name of the well (e.g. "A1 of Some Labware on 5")."""
        return self._display_name

    def get_name(self) -> str:
        """Get the name of the well (e.g. "A1")."""
        return self._name

    def get_column_name(self) -> str:
        """Get the column portion of the well name (e.g. "1")."""
        return self._column_name

    def get_row_name(self) -> str:
        """Get the row portion of the well name (e.g. "A")."""
        return self._row_name

    def get_max_volume(self) -> float:
        """Get the well's maximum liquid volume."""
        return self._geometry.max_volume

    def get_top(self, z_offset: float) -> Point:
        """Get the coordinate of the well's top, with an z-offset."""
        return self._geometry.top(z_offset)

    def get_bottom(self, z_offset: float) -> Point:
        """Get the coordinate of the well's bottom, with an z-offset."""
        return self._geometry.bottom(z_offset)

    def get_center(self) -> Point:
        """Get the coordinate of the well's center."""
        return self._geometry.center()

    def get_meniscus(self) -> Union[Point, SimulatedProbeResult]:
        """Get the coordinate of the well's center."""
        raise APIVersionError(api_element="Getting a meniscus")

    def load_liquid(
        self,
        liquid: Liquid,
        volume: float,
    ) -> None:
        """Load liquid into a well."""
        raise APIVersionError(api_element="Loading a liquid")

    def from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """Gets point in deck coordinates based on percentage of the radius of each axis."""
        return self._geometry.from_center_cartesian(x, y, z)

    def estimate_liquid_height_after_pipetting(
        self,
        mount: Mount | str,
        operation_volume: float,
    ) -> LiquidTrackingType:
        """Estimate what the liquid height will be after pipetting, without raising an error."""
        return 0.0

    def current_liquid_height(self) -> LiquidTrackingType:
        """Get the current liquid height."""
        return 0.0

    def get_liquid_volume(self) -> LiquidTrackingType:
        """Get the current well volume."""
        return 0.0

    def height_from_volume(self, volume: LiquidTrackingType) -> LiquidTrackingType:
        """Return the height in a well corresponding to a given volume."""
        return 0.0

    def volume_from_height(self, height: LiquidTrackingType) -> LiquidTrackingType:
        """Return the volume contained in a well at any height."""
        return 0.0

    # TODO(mc, 2022-10-28): is this used and/or necessary?
    def __repr__(self) -> str:
        """Use the well's display name as its repr."""
        return self.get_display_name()

    # TODO(mc, 2022-08-24): comparing only names seems insufficient
    def __eq__(self, other: object) -> bool:
        """Assume that if name is the same then it's the same well."""
        return isinstance(other, LegacyWellCore) and self.get_name() == other.get_name()



"""Seal tips to pipette command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union

from typing_extensions import Literal
from pydantic import Field, BaseModel

from opentrons_shared_data.errors.exceptions import PositionUnknownError

from opentrons.types import MountType
from opentrons.protocol_engine.types import MotorAxis
from ..resources import ModelUtils, ensure_ot3_hardware
from ..types import PickUpTipWellLocation, FluidKind, AspiratedFluid
from .pipetting_common import (
    PipetteIdMixin,
)
from .movement_common import (
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import Axis

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import (
        MovementHandler,
        TipHandler,
        GantryMover,
        PipettingHandler,
    )


SealPipetteToTipCommandType = Literal["sealPipetteToTip"]
_CAM_PREP_DISTANCE_DEFAULT = 8.25
_CAM_PRESS_DISTANCE_DEFAULT = 3.5
_CAM_EJECTOR_PUSH_MM_DEFAULT = 7.0
_PRESS_FIT_PREP_DISTANCE_DEFAULT = 0
_PRESS_FIT_PRESS_DISTANCE_DEFAULT = -11.0
_PRESS_FIT_EJECTOR_PUSH_MM_DEFAULT = 0
_SAFE_TOP_VOLUME = 1000


class TipPickUpParams(BaseModel):
    """Payload used to specify press-tip parameters for a seal command."""

    prepDistance: float = Field(
        default=0, description="The distance to move down to fit the tips on."
    )
    pressDistance: float = Field(
        default=0, description="The distance to press on tips."
    )
    ejectorPushMm: float = Field(
        default=0,
        description="The distance to back off to ensure that the tip presence sensors are not triggered.",
    )


class SealPipetteToTipParams(PipetteIdMixin):
    """Payload needed to seal resin tips to a pipette."""

    labwareId: str = Field(..., description="Identifier of labware to use.")
    wellName: str = Field(..., description="Name of well to use in labware.")
    wellLocation: PickUpTipWellLocation = Field(
        default_factory=PickUpTipWellLocation,
        description="Relative well location at which to pick up the tip.",
    )
    tipPickUpParams: Optional[TipPickUpParams] = Field(
        default=None, description="Specific parameters for "
    )


class SealPipetteToTipResult(DestinationPositionResult):
    """Result data from the execution of a SealPipetteToTip."""

    tipVolume: float = Field(
        0,
        description="Maximum volume of liquid that the picked up tip can hold, in µL.",
        ge=0,
    )

    tipLength: float = Field(
        0,
        description="The length of the tip in mm.",
        ge=0,
    )

    tipDiameter: float = Field(
        0,
        description="The diameter of the tip in mm.",
        ge=0,
    )


_ExecuteReturn = Union[
    SuccessData[SealPipetteToTipResult],
    DefinedErrorData[StallOrCollisionError],
]


class SealPipetteToTipImplementation(
    AbstractCommandImpl[SealPipetteToTipParams, _ExecuteReturn]
):
    """Seal pipette command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        model_utils: ModelUtils,
        movement: MovementHandler,
        hardware_api: HardwareControlAPI,
        gantry_mover: GantryMover,
        pipetting: PipettingHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._model_utils = model_utils
        self._movement = movement
        self._gantry_mover = gantry_mover
        self._pipetting = pipetting
        self._hardware_api = hardware_api

    async def relative_pickup_tip(
        self,
        tip_pick_up_params: TipPickUpParams,
        mount: MountType,
    ) -> None:
        """A relative press-fit pick up command using gantry moves."""
        prep_distance = tip_pick_up_params.prepDistance
        press_distance = tip_pick_up_params.pressDistance
        retract_distance = -1 * (press_distance) / 2

        mount_axis = MotorAxis.LEFT_Z if mount == MountType.LEFT else MotorAxis.RIGHT_Z
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        # TODO chb, 2025-01-29): Factor out the movement constants and relocate this logic into the hardware controller
        try:
            await self._gantry_mover.move_axes(
                axis_map={mount_axis: prep_distance},
                speed=10,
                relative_move=True,
                expect_stalls=True,
            )
        except PositionUnknownError:
            # if this happens it's from the get position after the move and we can ignore it
            pass

        await ot3_hardware_api.update_axis_position_estimations(
            self._gantry_mover.motor_axes_to_present_hardware_axes([mount_axis])
        )

        # Drive mount down for press-fit
        try:
            await self._gantry_mover.move_axes(
                axis_map={mount_axis: press_distance},
                speed=10.0,
                relative_move=True,
                expect_stalls=True,
            )
        except PositionUnknownError:
            # if this happens it's from the get position after the move and we can ignore it
            pass

        await ot3_hardware_api.update_axis_position_estimations(
            self._gantry_mover.motor_axes_to_present_hardware_axes([mount_axis])
        )

        try:
            await self._gantry_mover.move_axes(
                axis_map={mount_axis: retract_distance}, speed=5.5, relative_move=True
            )
        except PositionUnknownError:
            # if this happens it's from the get position after the move and we can ignore it
            pass

        await ot3_hardware_api.update_axis_position_estimations(
            self._gantry_mover.motor_axes_to_present_hardware_axes([mount_axis])
        )

    async def cam_action_relative_pickup_tip(
        self,
        tip_pick_up_params: TipPickUpParams,
        mount: MountType,
    ) -> None:
        """A cam action pick up command using gantry moves."""
        prep_distance = tip_pick_up_params.prepDistance
        press_distance = tip_pick_up_params.pressDistance
        ejector_push_mm = tip_pick_up_params.ejectorPushMm
        retract_distance = -1 * (prep_distance + press_distance)

        mount_axis = MotorAxis.LEFT_Z if mount == MountType.LEFT else MotorAxis.RIGHT_Z

        # TODO chb, 2025-01-29): Factor out the movement constants and relocate this logic into the hardware controller
        await self._gantry_mover.move_axes(
            axis_map={mount_axis: -6}, speed=10, relative_move=True
        )

        # Drive Q down 3mm at fast speed - look into the pick up tip fuinction to find slow and fast: 10.0
        await self._gantry_mover.move_axes(
            axis_map={MotorAxis.AXIS_96_CHANNEL_CAM: prep_distance},
            speed=10.0,
            relative_move=True,
        )
        # 2.8mm at slow speed - cam action pickup speed: 5.5
        await self._gantry_mover.move_axes(
            axis_map={MotorAxis.AXIS_96_CHANNEL_CAM: press_distance},
            speed=5.5,
            relative_move=True,
        )
        # retract cam : 11.05
        await self._gantry_mover.move_axes(
            axis_map={MotorAxis.AXIS_96_CHANNEL_CAM: retract_distance},
            speed=5.5,
            relative_move=True,
        )

        # Lower tip presence
        await self._gantry_mover.move_axes(
            axis_map={mount_axis: 2}, speed=10, relative_move=True
        )
        await self._gantry_mover.move_axes(
            axis_map={MotorAxis.AXIS_96_CHANNEL_CAM: ejector_push_mm},
            speed=5.5,
            relative_move=True,
        )
        await self._gantry_mover.move_axes(
            axis_map={MotorAxis.AXIS_96_CHANNEL_CAM: -1 * ejector_push_mm},
            speed=5.5,
            relative_move=True,
        )

    async def execute(
        self, params: SealPipetteToTipParams
    ) -> Union[SuccessData[SealPipetteToTipResult], _ExecuteReturn]:
        """Move to and pick up a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        well_location = self._state_view.geometry.convert_pick_up_tip_well_location(
            well_location=params.wellLocation
        )
        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result

        # Aspirate to move plunger to a maximum volume position per pipette type
        tip_geometry = self._state_view.geometry.get_nominal_tip_geometry(
            pipette_id, labware_id, well_name
        )
        if self._state_view.pipettes.get_mount(pipette_id) == MountType.LEFT:
            await self._hardware_api.home(axes=[Axis.P_L])
        else:
            await self._hardware_api.home(axes=[Axis.P_R])

        # Begin relative pickup steps for the resin tips

        channels = self._state_view.pipettes.get_active_channels(pipette_id)
        mount = self._state_view.pipettes.get_mount(pipette_id)
        tip_pick_up_params = params.tipPickUpParams

        if channels == 96:
            if tip_pick_up_params is None:
                tip_pick_up_params = TipPickUpParams(
                    prepDistance=_CAM_PREP_DISTANCE_DEFAULT,
                    pressDistance=_CAM_PRESS_DISTANCE_DEFAULT,
                    ejectorPushMm=_CAM_EJECTOR_PUSH_MM_DEFAULT,
                )
            await self.cam_action_relative_pickup_tip(
                tip_pick_up_params=tip_pick_up_params,
                mount=mount,
            )
        else:
            if tip_pick_up_params is None:
                tip_pick_up_params = TipPickUpParams(
                    prepDistance=_PRESS_FIT_PREP_DISTANCE_DEFAULT,
                    pressDistance=_PRESS_FIT_PRESS_DISTANCE_DEFAULT,
                    ejectorPushMm=_PRESS_FIT_EJECTOR_PUSH_MM_DEFAULT,
                )
            await self.relative_pickup_tip(
                tip_pick_up_params=tip_pick_up_params,
                mount=mount,
            )

        # cache_tip
        if self._state_view.config.use_virtual_pipettes is False:
            self._tip_handler.cache_tip(pipette_id, tip_geometry)
            hw_instr = self._hardware_api.hardware_instruments[mount.to_hw_mount()]
            if hw_instr is not None:
                hw_instr.set_current_volume(_SAFE_TOP_VOLUME)

        state_update = move_result.state_update.update_pipette_tip_state(
            pipette_id=pipette_id,
            tip_geometry=tip_geometry,
        ).set_fluid_aspirated(
            pipette_id=pipette_id,
            fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=_SAFE_TOP_VOLUME),
        )
        return SuccessData(
            public=SealPipetteToTipResult(
                tipVolume=tip_geometry.volume,
                tipLength=tip_geometry.length,
                tipDiameter=tip_geometry.diameter,
                position=move_result.public.position,
            ),
            state_update=state_update,
        )


class SealPipetteToTip(
    BaseCommand[
        SealPipetteToTipParams,
        SealPipetteToTipResult,
        StallOrCollisionError,
    ]
):
    """Seal tip command model."""

    commandType: SealPipetteToTipCommandType = "sealPipetteToTip"
    params: SealPipetteToTipParams
    result: Optional[SealPipetteToTipResult] = None

    _ImplementationCls: Type[
        SealPipetteToTipImplementation
    ] = SealPipetteToTipImplementation


class SealPipetteToTipCreate(BaseCommandCreate[SealPipetteToTipParams]):
    """Seal tip command creation request model."""

    commandType: SealPipetteToTipCommandType = "sealPipetteToTip"
    params: SealPipetteToTipParams

    _CommandCls: Type[SealPipetteToTip] = SealPipetteToTip



"""Command models to engage a user to empty a Flex Stacker."""

from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING, Annotated
from typing_extensions import Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors import (
    ErrorOccurrence,
)
from ...errors.exceptions import FlexStackerLabwarePoolNotYetDefinedError
from ...state import update_types
from ...types import StackerFillEmptyStrategy
from opentrons.calibration_storage.helpers import uri_from_details

if TYPE_CHECKING:
    from ...state.state import StateView
    from ...execution import RunControlHandler

FillCommandType = Literal["flexStacker/fill"]


class FillParams(BaseModel):
    """The parameters defining how a stacker should be filled."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")

    strategy: StackerFillEmptyStrategy = Field(
        ...,
        description=(
            "How to fill the stacker. "
            "If manualWithPause, pause the protocol until the client sends an interaction, and apply "
            "the new specified count thereafter. If logical, do not pause but immediately apply the "
            "specified count."
        ),
    )

    message: str | SkipJsonSchema[None] = Field(
        None,
        description="The message to display on connected clients during a manualWithPause strategy fill.",
    )

    count: Optional[Annotated[int, Field(ge=1)]] = Field(
        None,
        description=(
            "How full the labware pool should now be. If None, default to the maximum amount "
            "of the currently-configured labware the pool can hold. "
            "If this number is larger than the maximum the pool can hold, it will be clamped to "
            "the maximum. If this number is smaller than the current amount of labware the pool "
            "holds, it will be clamped to that minimum. Do not use the value in the parameters as "
            "an outside observer; instead, use the count value from the results."
        ),
    )


class FillResult(BaseModel):
    """Result data from a stacker fill command."""

    count: int = Field(
        ..., description="The new amount of labware stored in the stacker labware pool."
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )


class FillImpl(AbstractCommandImpl[FillParams, SuccessData[FillResult]]):
    """Implementation of a stacker fill command."""

    def __init__(
        self, state_view: StateView, run_control: RunControlHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._run_control = run_control

    async def execute(self, params: FillParams) -> SuccessData[FillResult]:
        """Execute the stacker fill command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        if stacker_state.pool_primary_definition is None:
            location = self._state_view.modules.get_location(params.moduleId)
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message=f"The Flex Stacker in {location} has not been configured yet and cannot be filled."
            )

        count = (
            params.count if params.count is not None else stacker_state.max_pool_count
        )
        new_count = min(
            stacker_state.max_pool_count, max(stacker_state.pool_count, count)
        )

        state_update = (
            update_types.StateUpdate().update_flex_stacker_labware_pool_count(
                params.moduleId, new_count
            )
        )

        if params.strategy == StackerFillEmptyStrategy.MANUAL_WITH_PAUSE:
            await self._run_control.wait_for_resume()

        if stacker_state.pool_primary_definition is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                "The Primary Labware must be defined in the stacker pool."
            )

        return SuccessData(
            public=FillResult(
                count=new_count,
                primaryLabwareURI=uri_from_details(
                    stacker_state.pool_primary_definition.namespace,
                    stacker_state.pool_primary_definition.parameters.loadName,
                    stacker_state.pool_primary_definition.version,
                ),
                adapterLabwareURI=uri_from_details(
                    stacker_state.pool_adapter_definition.namespace,
                    stacker_state.pool_adapter_definition.parameters.loadName,
                    stacker_state.pool_adapter_definition.version,
                )
                if stacker_state.pool_adapter_definition is not None
                else None,
                lidLabwareURI=uri_from_details(
                    stacker_state.pool_lid_definition.namespace,
                    stacker_state.pool_lid_definition.parameters.loadName,
                    stacker_state.pool_lid_definition.version,
                )
                if stacker_state.pool_lid_definition is not None
                else None,
            ),
            state_update=state_update,
        )


class Fill(BaseCommand[FillParams, FillResult, ErrorOccurrence]):
    """A command to fill the Flex Stacker with labware."""

    commandType: FillCommandType = "flexStacker/fill"
    params: FillParams
    result: Optional[FillResult] = None

    _ImplementationCls: Type[FillImpl] = FillImpl


class FillCreate(BaseCommandCreate[FillParams]):
    """A request to execute a Flex Stacker fill command."""

    commandType: FillCommandType = "flexStacker/fill"
    params: FillParams

    _CommandCls: Type[Fill] = Fill



import json
import logging
from pydantic import ValidationError
from typing import Optional, Union, no_type_check
from dataclasses import asdict

from opentrons import config
from opentrons.util.helpers import utc_now


from .. import file_operators as io, types as local_types

from .models import v1

log = logging.getLogger(__name__)


# Delete Belt Calibration


@no_type_check
def delete_robot_belt_attitude() -> None:
    """
    Delete the robot belt attitude calibration.
    """
    gantry_path = (
        config.get_opentrons_path("robot_calibration_dir") / "belt_calibration.json"
    )

    io.delete_file(gantry_path)


# Save Belt Calibration


@no_type_check
def save_robot_belt_attitude(
    transform: local_types.AttitudeMatrix,
    pip_id: Optional[str],
    source: Optional[local_types.SourceType] = None,
    cal_status: Optional[
        Union[local_types.CalibrationStatus, local_types.CalibrationStatus]
    ] = None,
) -> None:
    robot_dir = config.get_opentrons_path("robot_calibration_dir")

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    gantry_calibration = v1.BeltCalibrationModel(
        attitude=transform,
        pipetteCalibratedWith=pip_id,
        lastModified=utc_now(),
        source=source or local_types.SourceType.user,
        status=cal_status_model,
    )
    # convert to schema + validate json conversion
    io.save_to_file(robot_dir, "belt_calibration", gantry_calibration)


# Get Belt Calibration


@no_type_check
def get_robot_belt_attitude() -> Optional[v1.BeltCalibrationModel]:
    belt_calibration_path = (
        config.get_opentrons_path("robot_calibration_dir") / "belt_calibration.json"
    )
    try:
        return v1.BeltCalibrationModel(**io.read_cal_file(belt_calibration_path))
    except FileNotFoundError:
        log.warning("Belt calibration not found.")
        pass
    except (json.JSONDecodeError, ValidationError):
        log.warning(
            "Belt calibration is malformed. Please factory reset your calibrations.",
            exc_info=True,
        )
    return None



"""Unseal tip from pipette command request, result, and implementation models."""

from __future__ import annotations

from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type, Final
from typing_extensions import Literal

from opentrons.protocol_engine.resources.model_utils import ModelUtils

from ..types import DropTipWellLocation
from .pipetting_common import (
    PipetteIdMixin,
)
from .movement_common import (
    DestinationPositionResult,
    move_to_well,
    StallOrCollisionError,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import MovementHandler, TipHandler, GantryMover


UnsealPipetteFromTipCommandType = Literal["unsealPipetteFromTip"]


class UnsealPipetteFromTipParams(PipetteIdMixin):
    """Payload required to drop a tip in a specific well."""

    labwareId: str = Field(..., description="Identifier of labware to use.")
    wellName: str = Field(..., description="Name of well to use in labware.")
    wellLocation: DropTipWellLocation = Field(
        default_factory=DropTipWellLocation,
        description="Relative well location at which to drop the tip.",
    )


class UnsealPipetteFromTipResult(DestinationPositionResult):
    """Result data from the execution of a DropTip command."""

    pass


_ExecuteReturn = (
    SuccessData[UnsealPipetteFromTipResult] | DefinedErrorData[StallOrCollisionError]
)

CUSTOM_TIP_LENGTH_MARGIN: Final = 10


class UnsealPipetteFromTipImplementation(
    AbstractCommandImpl[UnsealPipetteFromTipParams, _ExecuteReturn]
):
    """Drop tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        movement: MovementHandler,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._movement_handler = movement
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: UnsealPipetteFromTipParams) -> _ExecuteReturn:
        """Move to and drop a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        well_location = params.wellLocation

        tip_geometry = self._state_view.geometry.get_nominal_tip_geometry(
            pipette_id, labware_id, well_name
        )

        is_partially_configured = self._state_view.pipettes.get_is_partially_configured(
            pipette_id=pipette_id
        )
        tip_drop_location = self._state_view.geometry.get_checked_tip_drop_location(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_location=well_location,
            partially_configured=is_partially_configured,
            override_default_offset=-(tip_geometry.length - CUSTOM_TIP_LENGTH_MARGIN),
        )

        move_result = await move_to_well(
            movement=self._movement_handler,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=tip_drop_location,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result

        await self._tip_handler.drop_tip(
            pipette_id=pipette_id,
            home_after=None,
            do_not_ignore_tip_presence=False,
            ignore_plunger=True,
        )

        return SuccessData(
            public=UnsealPipetteFromTipResult(position=move_result.public.position),
            state_update=move_result.state_update.set_fluid_unknown(
                pipette_id=pipette_id
            ).update_pipette_tip_state(pipette_id=params.pipetteId, tip_geometry=None),
        )


class UnsealPipetteFromTip(
    BaseCommand[
        UnsealPipetteFromTipParams, UnsealPipetteFromTipResult, StallOrCollisionError
    ]
):
    """Unseal pipette command model."""

    commandType: UnsealPipetteFromTipCommandType = "unsealPipetteFromTip"
    params: UnsealPipetteFromTipParams
    result: Optional[UnsealPipetteFromTipResult] = None

    _ImplementationCls: Type[
        UnsealPipetteFromTipImplementation
    ] = UnsealPipetteFromTipImplementation


class UnsealPipetteFromTipCreate(BaseCommandCreate[UnsealPipetteFromTipParams]):
    """Unseal pipette command creation request model."""

    commandType: UnsealPipetteFromTipCommandType = "unsealPipetteFromTip"
    params: UnsealPipetteFromTipParams

    _CommandCls: Type[UnsealPipetteFromTip] = UnsealPipetteFromTip



"""Get next tip command request, result, and implementation models."""

from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Any, Optional, Type, List, Literal, Union

from pydantic.json_schema import SkipJsonSchema

from opentrons.types import NozzleConfigurationType

from ..errors import ErrorOccurrence
from ..types import NextTipInfo, NoTipAvailable, NoTipReason
from .pipetting_common import PipetteIdMixin

from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)

if TYPE_CHECKING:
    from ..state.state import StateView


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


GetNextTipCommandType = Literal["getNextTip"]


class GetNextTipParams(PipetteIdMixin):
    """Payload needed to resolve the next available tip."""

    labwareIds: List[str] = Field(
        ...,
        description="Labware ID(s) of tip racks to resolve next available tip(s) from"
        " Labware IDs will be resolved sequentially",
    )
    startingTipWell: str | SkipJsonSchema[None] = Field(
        None,
        description="Name of starting tip rack 'well'."
        " This only applies to the first tip rack in the list provided in labwareIDs",
        json_schema_extra=_remove_default,
    )


class GetNextTipResult(BaseModel):
    """Result data from the execution of a GetNextTip."""

    nextTipInfo: Union[NextTipInfo, NoTipAvailable] = Field(
        ...,
        description="Labware ID and well name of next available tip for a pipette,"
        " or information why no tip could be resolved.",
    )


class GetNextTipImplementation(
    AbstractCommandImpl[GetNextTipParams, SuccessData[GetNextTipResult]]
):
    """Get next tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view

    async def execute(self, params: GetNextTipParams) -> SuccessData[GetNextTipResult]:
        """Get the next available tip for the requested pipette."""
        pipette_id = params.pipetteId
        starting_tip_name = params.startingTipWell

        num_tips = self._state_view.pipettes.get_active_channels(pipette_id)
        nozzle_map = self._state_view.pipettes.get_nozzle_configuration(pipette_id)

        if (
            starting_tip_name is not None
            and nozzle_map.configuration != NozzleConfigurationType.FULL
        ):
            # This is to match the behavior found in PAPI, but also because we don't have logic to automatically find
            # the next tip with partial configuration and a starting tip. This will never work for a 96-channel due to
            # x-axis overlap, but could eventually work with 8-channel if we better define starting tip USED or CLEAN
            # state when starting a protocol to prevent accidental tip pick-up with starting non-full tip racks.
            return SuccessData(
                public=GetNextTipResult(
                    nextTipInfo=NoTipAvailable(
                        noTipReason=NoTipReason.STARTING_TIP_WITH_PARTIAL,
                        message="Cannot automatically resolve next tip with starting tip and partial tip configuration.",
                    )
                )
            )

        next_tip: Union[NextTipInfo, NoTipAvailable]
        for labware_id in params.labwareIds:
            well_name = self._state_view.tips.get_next_tip(
                labware_id=labware_id,
                num_tips=num_tips,
                starting_tip_name=starting_tip_name,
                nozzle_map=nozzle_map,
            )
            if well_name is not None:
                next_tip = NextTipInfo(labwareId=labware_id, tipStartingWell=well_name)
                break
            # After the first tip rack is exhausted, starting tip no longer applies
            starting_tip_name = None
        else:
            next_tip = NoTipAvailable(
                noTipReason=NoTipReason.NO_AVAILABLE_TIPS,
                message="No available tips for given pipette, nozzle configuration and provided tip racks.",
            )

        return SuccessData(public=GetNextTipResult(nextTipInfo=next_tip))


class GetNextTip(BaseCommand[GetNextTipParams, GetNextTipResult, ErrorOccurrence]):
    """Get next tip command model."""

    commandType: GetNextTipCommandType = "getNextTip"
    params: GetNextTipParams
    result: Optional[GetNextTipResult] = None

    _ImplementationCls: Type[GetNextTipImplementation] = GetNextTipImplementation


class GetNextTipCreate(BaseCommandCreate[GetNextTipParams]):
    """Get next tip command creation request model."""

    commandType: GetNextTipCommandType = "getNextTip"
    params: GetNextTipParams

    _CommandCls: Type[GetNextTip] = GetNextTip



from enum import Enum


class ModuleType(str, Enum):
    """Module type enumeration."""

    Magnetic = "magdeck"
    Temperature = "tempdeck"
    Thermocycler = "thermocycler"
    Heatershaker = "heatershaker"



"""Translation of JSON protocol commands into ProtocolEngine commands."""

from typing import List, Union, Iterator
from pydantic import ValidationError as PydanticValidationError, TypeAdapter

from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.protocol.models import (
    ProtocolSchemaV6,
    protocol_schema_v6,
    ProtocolSchemaV7,
    protocol_schema_v7,
    ProtocolSchemaV8,
    protocol_schema_v8,
    Location,
    #    CommandSchemaId,
)
from opentrons_shared_data import command as command_schema
from opentrons_shared_data.errors.exceptions import InvalidProtocolData, PythonException

from opentrons.types import MountType
from opentrons.protocol_engine import (
    commands as pe_commands,
    LoadableLabwareLocation,
    ModuleModel,
    DeckSlotLocation,
    Liquid,
)
from opentrons.protocol_engine.types import HexColor, CommandAnnotation


class CommandTranslatorError(Exception):
    """An error raised to indicate an internal programmer error."""

    pass


# Each time a TypeAdapter is instantiated, it will construct a new validator and
# serializer. To improve performance, TypeAdapters are instantiated once.
# See https://docs.pydantic.dev/latest/concepts/performance/#typeadapter-instantiated-once
LabwareLocationAdapter: TypeAdapter[LoadableLabwareLocation] = TypeAdapter(
    LoadableLabwareLocation
)
CommandAnnotationAdapter: TypeAdapter[CommandAnnotation] = TypeAdapter(
    CommandAnnotation
)


def _translate_labware_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.LoadLabwareCreate:
    labware_id = command.params.labwareId
    # v6 data model supports all commands and therefor most props are optional.
    # load labware command must contain labware_id and definition_id.
    assert labware_id is not None
    definition_id = protocol.labware[labware_id].definitionId
    assert definition_id is not None

    location = command.params.location
    labware_command = pe_commands.LoadLabwareCreate(
        params=pe_commands.LoadLabwareParams(
            labwareId=command.params.labwareId,
            displayName=protocol.labware[labware_id].displayName,
            version=protocol.labwareDefinitions[definition_id].version,
            namespace=protocol.labwareDefinitions[definition_id].namespace,
            loadName=protocol.labwareDefinitions[definition_id].parameters.loadName,
            location=LabwareLocationAdapter.validate_python(
                location.model_dump() if isinstance(location, Location) else location
            ),
        ),
        key=command.key,
    )
    return labware_command


def _translate_v7_labware_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.LoadLabwareCreate:
    labware_id = command.params.labwareId
    # v7 data model supports all commands and therefor most props are optional.
    # load labware command must contain labware_id and definition_id.
    assert labware_id is not None
    assert command.params.version is not None
    assert command.params.namespace is not None
    assert command.params.loadName is not None

    location = command.params.location
    labware_command = pe_commands.LoadLabwareCreate(
        params=pe_commands.LoadLabwareParams(
            labwareId=command.params.labwareId,
            displayName=command.params.displayName,
            version=command.params.version,
            namespace=command.params.namespace,
            loadName=command.params.loadName,
            location=LabwareLocationAdapter.validate_python(
                location.model_dump() if isinstance(location, Location) else location
            ),
        ),
        key=command.key,
    )
    return labware_command


def _translate_module_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.CommandCreate:
    module_id = command.params.moduleId
    modules = protocol.modules
    # v6 data model supports all commands and therefor most props are optional.
    # load module command must contain module_id. modules cannot be None.
    assert module_id is not None
    assert modules is not None

    location = command.params.location
    translated_obj = pe_commands.LoadModuleCreate(
        params=pe_commands.LoadModuleParams(
            model=ModuleModel(modules[module_id].model),
            location=DeckSlotLocation.model_validate(
                location.model_dump() if isinstance(location, Location) else location
            ),
            moduleId=command.params.moduleId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_v7_module_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.CommandCreate:
    module_id = command.params.moduleId
    # v7 data model supports all commands and therefor most props are optional.
    # load module command must contain module_id. modules cannot be None.
    assert module_id is not None
    assert command.params.model is not None
    location = command.params.location
    translated_obj = pe_commands.LoadModuleCreate(
        params=pe_commands.LoadModuleParams(
            model=ModuleModel(command.params.model),
            location=DeckSlotLocation.model_validate(
                location.model_dump() if isinstance(location, Location) else location
            ),
            moduleId=command.params.moduleId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_pipette_command(
    protocol: ProtocolSchemaV6,
    command: protocol_schema_v6.Command,
) -> pe_commands.CommandCreate:
    pipette_id = command.params.pipetteId
    # v6 data model supports all commands and therefor most props are optional.
    # load pipette command must contain pipette_id.
    assert pipette_id is not None
    translated_obj = pe_commands.LoadPipetteCreate(
        params=pe_commands.LoadPipetteParams(
            pipetteName=PipetteNameType(protocol.pipettes[pipette_id].name),
            mount=MountType(command.params.mount),
            pipetteId=command.params.pipetteId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_v7_pipette_command(
    command: protocol_schema_v7.Command,
) -> pe_commands.CommandCreate:
    pipette_id = command.params.pipetteId
    # v7 data model supports all commands and therefor most props are optional.
    # load pipette command must contain pipette_id.
    assert pipette_id is not None
    assert command.params.pipetteName is not None
    translated_obj = pe_commands.LoadPipetteCreate(
        params=pe_commands.LoadPipetteParams(
            pipetteName=PipetteNameType(command.params.pipetteName),
            mount=MountType(command.params.mount),
            pipetteId=command.params.pipetteId,
        ),
        key=command.key,
    )
    return translated_obj


def _translate_simple_command(
    command: Union[
        protocol_schema_v6.Command,
        protocol_schema_v7.Command,
        protocol_schema_v8.Command,
    ],
) -> pe_commands.CommandCreate:
    dict_command = command.model_dump(exclude_none=True)

    # map deprecated `delay` commands to `waitForResume` / `waitForDuration`
    if dict_command["commandType"] == "delay":
        if "waitForResume" in dict_command["params"]:
            dict_command["commandType"] = "waitForResume"
        else:
            dict_command["commandType"] = "waitForDuration"

    return pe_commands.CommandCreateAdapter.validate_python(dict_command)


class JsonTranslator:
    """Class that translates commands/liquids from PD/JSON to ProtocolEngine."""

    def translate_liquids(
        self, protocol: Union[ProtocolSchemaV6, ProtocolSchemaV7, ProtocolSchemaV8]
    ) -> List[Liquid]:
        """Takes json protocol v6 and translates liquids->protocol engine liquids."""
        protocol_liquids = protocol.liquids or {}

        return [
            Liquid(
                id=liquid_id,
                displayName=liquid.displayName,
                description=liquid.description,
                displayColor=HexColor(liquid.displayColor)
                if liquid.displayColor is not None
                else None,
            )
            for liquid_id, liquid in protocol_liquids.items()
        ]

    def translate_commands(
        self,
        protocol: Union[ProtocolSchemaV8, ProtocolSchemaV7, ProtocolSchemaV6],
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol and translates commands->protocol engine commands."""
        if isinstance(protocol, ProtocolSchemaV6):
            return self._translate_v6_commands(protocol)
        elif isinstance(protocol, ProtocolSchemaV7):
            return self._translate_v7_commands(protocol)
        else:
            return self._translate_v8_commands(protocol)

    def _translate_v6_commands(
        self,
        protocol: ProtocolSchemaV6,
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol v6 and translates commands->protocol engine commands."""
        commands_list: List[pe_commands.CommandCreate] = []
        for command in protocol.commands:
            if command.commandType == "loadPipette":
                translated_obj = _translate_pipette_command(protocol, command)
            elif command.commandType == "loadModule":
                translated_obj = _translate_module_command(protocol, command)
            elif command.commandType == "loadLabware":
                translated_obj = _translate_labware_command(protocol, command)
            else:
                translated_obj = _translate_simple_command(command)
            commands_list.append(translated_obj)
        return commands_list

    def _translate_v7_commands(
        self,
        protocol: Union[ProtocolSchemaV7],
    ) -> List[pe_commands.CommandCreate]:
        """Takes json protocol v7 and translates commands->protocol engine commands."""
        commands_list: List[pe_commands.CommandCreate] = []
        for command in protocol.commands:
            translated_obj = _translate_simple_command(command)
            commands_list.append(translated_obj)
        return commands_list

    def _translate_v8_commands(
        self, protocol: ProtocolSchemaV8
    ) -> List[pe_commands.CommandCreate]:
        """Translate commands in json protocol schema v8, which might be of different command schemas."""
        command_schema_ref = protocol.commandSchemaId.value

        # these calls will raise if the command schema version is invalid or unknown
        command_schema_version = command_schema.schema_version_from_ref(
            command_schema_ref
        )
        command_schema_string = command_schema.load_schema_string(  # noqa: F841
            command_schema_version
        )

        def translate_all_commands() -> Iterator[pe_commands.CommandCreate]:
            for command in protocol.commands:
                try:
                    yield _translate_simple_command(command)
                except PydanticValidationError as pve:
                    raise InvalidProtocolData(
                        message=(
                            "The protocol is invalid because it contains an unknown or malformed command, "
                            f'"{command.commandType}".'
                        ),
                        detail={"kind": "invalid-command"},
                        wrapping=[PythonException(pve)],
                    )

        return list(translate_all_commands())

    def translate_command_annotations(
        self,
        protocol: Union[ProtocolSchemaV8, ProtocolSchemaV7, ProtocolSchemaV6],
    ) -> List[CommandAnnotation]:
        """Translate command annotations in json protocol schema v8."""
        if isinstance(protocol, (ProtocolSchemaV6, ProtocolSchemaV7)):
            return []
        else:
            command_annotations: List[CommandAnnotation] = [
                CommandAnnotationAdapter.validate_python(
                    command_annotation.model_dump(),
                )
                for command_annotation in protocol.commandAnnotations
            ]
            return command_annotations



from .driver import HeaterShakerDriver
from .simulator import SimulatingDriver
from .abstract import AbstractHeaterShakerDriver

__all__ = ["HeaterShakerDriver", "SimulatingDriver", "AbstractHeaterShakerDriver"]



import json
import logging
import os
import sys
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Mapping,
    Tuple,
    Union,
    Optional,
    NamedTuple,
    cast,
    List,
)

from opentrons.config import CONFIG
from opentrons_shared_data.robot.types import RobotTypeEnum

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)


SettingsMap = Dict[str, Optional[bool]]


class SettingException(Exception):
    def __init__(self, message: str, error: str) -> None:
        super(Exception, self).__init__(message)
        self.error = error


class SettingsData(NamedTuple):
    settings_map: SettingsMap
    version: int


class SettingDefinition:
    def __init__(
        self,
        _id: str,
        title: str,
        description: str,
        robot_type: List[RobotTypeEnum],
        old_id: Optional[str] = None,
        restart_required: bool = False,
        show_if: Optional[Tuple[str, bool]] = None,
        internal_only: bool = False,
        default_true_on_robot_types: Optional[List[RobotTypeEnum]] = None,
    ):
        self.id = _id
        #: The id of the setting for programmatic access through
        #: get_adv_setting
        self.old_id = old_id
        #: the old id before migration, if any
        self.title = title
        #: User facing title
        self.description = description
        #: User facing description
        self.restart_required = restart_required
        #: True if the user must restart
        self.show_if = show_if
        #: A tuple of (other setting id, setting value) that must match reality
        #: to show this setting in http endpoints
        self.robot_type = robot_type
        #: A list of RobotTypeEnums that are compatible with this feature flag.
        self.internal_only = internal_only
        #: A flag determining whether this setting is user-facing.
        self.default_true_on_robot_types = default_true_on_robot_types or []
        #: Robot types for which null/unset means the setting is activated

    def __repr__(self) -> str:
        return "{}: {}".format(self.__class__, self.id)

    def should_show(self) -> bool:
        """
        Use show_if attribute to determine if setting should be presented
        to users
        """
        if not self.show_if:
            return True
        return (
            get_setting_with_env_overload(self.show_if[0], self.robot_type[0])
            == self.show_if[1]
        )

    async def on_change(self, value: Optional[bool]) -> None:
        """
        An opportunity for side effects as a result of change a setting
        value
        """
        if self.restart_required:
            set_restart_required()


class Setting(NamedTuple):
    value: Optional[bool]
    definition: SettingDefinition


# If you add or remove any settings here BE SURE TO ADD A MIGRATION below.
# You will also need to update the migration tests in:
# api/tests/opentrons/config/test_advanced_settings_migration.py
settings = [
    SettingDefinition(
        _id="shortFixedTrash",
        old_id="short-fixed-trash",
        title="Short (55mm) fixed trash",
        description="Trash box is 55mm tall (rather than the 77mm default)",
        robot_type=[RobotTypeEnum.OT2],
    ),
    SettingDefinition(
        _id="deckCalibrationDots",
        old_id="dots-deck-type",
        title="Deck calibration to dots",
        description="Perform deck calibration to dots rather than crosses, for"
        " robots that do not have crosses etched on the deck",
        robot_type=[RobotTypeEnum.OT2],
    ),
    SettingDefinition(
        _id="disableHomeOnBoot",
        old_id="disable-home-on-boot",
        title="Disable home on boot",
        description="Prevent robot from homing motors on boot",
        robot_type=[RobotTypeEnum.OT2, RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="useOldAspirationFunctions",
        title="Use older aspirate behavior",
        description="Aspirate with the less accurate volumetric calibrations"
        " that were used before version 3.7.0. Use this if you"
        " need consistency with pre-v3.7.0 results. This only"
        " affects GEN1 P10S, P10M, P50S, P50M, and P300S pipettes.",
        robot_type=[RobotTypeEnum.OT2],
    ),
    SettingDefinition(
        _id="enableDoorSafetySwitch",
        title="Enable robot door safety switch",
        description="Automatically pause protocols when robot door opens. "
        "Opening the robot door during a run will "
        "pause your robot only after it has completed its "
        "current motion.",
        robot_type=[RobotTypeEnum.OT2],
        default_true_on_robot_types=[RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="enableOT3HardwareController",
        title="Enable experimental OT-3 hardware controller",
        description=(
            "Do not enable. This is an Opentrons-internal setting to test "
            "new hardware."
        ),
        restart_required=True,
        robot_type=[RobotTypeEnum.FLEX],
        internal_only=True,
    ),
    SettingDefinition(
        _id="rearPanelIntegration",
        title="Enable robots with the new usb connected rear-panel board.",
        description="This is an Opentrons-internal setting to test new rear-panel.",
        robot_type=[RobotTypeEnum.FLEX],
        internal_only=True,
    ),
    SettingDefinition(
        _id="disableStallDetection",
        title="Disable stall detection on the Flex.",
        description="This is an Opentrons-internal setting for hardware-testing.",
        robot_type=[RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="disableStatusBar",
        title="Disable the LED status bar on the Flex.",
        description="This setting disables the LED status bar on the Flex.",
        robot_type=[RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="estopNotRequired",
        title="If enabled, the Flex gantry can move with no estop attached.",
        description="This setting allows the gantry on the Flex to move with no estop attached.",
        show_if=(
            "estopNotRequired",
            True,
        ),  # Configured so this only shows if it has been set by a user
        robot_type=[RobotTypeEnum.FLEX],
        internal_only=True,
    ),
    SettingDefinition(
        _id="disableOverpressureDetection",
        title="Disable Flex pipette pressure sensing.",
        description="When this setting is on, Flex will continue its activities regardless of pressure changes inside the pipette. Do not turn this setting on unless you are intentionally causing pressures over 8 kPa inside the pipette air channel.",
        robot_type=[RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="enableErrorRecoveryExperiments",
        title="Enable error recovery experiments",
        description=(
            "Do not enable."
            " This is an Opentrons internal setting to enable additional,"
            " in-development error recovery features."
        ),
        robot_type=[RobotTypeEnum.FLEX],
        internal_only=True,
    ),
    SettingDefinition(
        _id="enableOEMMode",
        title="Enable OEM Mode",
        description="This setting anonymizes Opentrons branding in the ODD app.",
        robot_type=[RobotTypeEnum.FLEX],
    ),
    SettingDefinition(
        _id="enablePerformanceMetrics",
        title="Enable performance metrics",
        description=(
            "Do not enable."
            " This is an Opentrons internal setting to collect performance metrics."
            " Do not turn this on unless you are playing with the performance metrics system."
        ),
        robot_type=[RobotTypeEnum.OT2, RobotTypeEnum.FLEX],
        internal_only=True,
    ),
]


settings_by_id: Dict[str, SettingDefinition] = {s.id: s for s in settings}
settings_by_old_id: Dict[str, SettingDefinition] = {
    s.old_id: s for s in settings if s.old_id
}


def get_setting_definition(setting_id: str) -> Optional[SettingDefinition]:
    clean = _clean_id(setting_id)
    for setting in settings:
        if setting.id == clean:
            return setting
    return None


def get_adv_setting(setting: str, robot_type: RobotTypeEnum) -> Optional[Setting]:
    setting = _clean_id(setting)
    s = get_all_adv_settings(robot_type, include_internal=True)
    return s.get(setting, None)


def _filtered_key(
    definition: SettingDefinition,
    robot_type: RobotTypeEnum,
    include_internal: bool,
) -> bool:
    if include_internal:
        return robot_type in definition.robot_type
    else:
        return robot_type in definition.robot_type and not definition.internal_only


@lru_cache(maxsize=1)
def get_all_adv_settings(
    robot_type: RobotTypeEnum, include_internal: bool = False
) -> Dict[str, Setting]:
    """Get all the advanced setting values and definitions"""
    settings_file = CONFIG["feature_flags_file"]

    values, _ = _read_settings_file(settings_file)

    return {
        key: Setting(value=value, definition=settings_by_id[key])
        for key, value in values.items()
        if key in settings_by_id
        and _filtered_key(settings_by_id[key], robot_type, include_internal)
    }


async def set_adv_setting(_id: str, value: Optional[bool]) -> None:
    _id = _clean_id(_id)
    settings_file = CONFIG["feature_flags_file"]
    setting_data = _read_settings_file(settings_file)
    if _id not in setting_data.settings_map:
        raise ValueError(f"{_id} is not recognized")
    # Side effecting
    await settings_by_id[_id].on_change(value)

    setting_data.settings_map[_id] = value
    _write_settings_file(setting_data.settings_map, setting_data.version, settings_file)
    # Clear the lru cache
    get_all_adv_settings.cache_clear()


def _clean_id(_id: str) -> str:
    if _id in settings_by_old_id.keys():
        _id = settings_by_old_id[_id].id
    return _id


def _read_json_file(path: Union[str, "Path"]) -> Dict[str, Any]:
    try:
        with open(path, "r") as fd:
            data = json.load(fd)
    except FileNotFoundError:
        data = {}
    except json.JSONDecodeError as e:
        sys.stderr.write(f"Could not load advanced settings file {path}: {e}\n")
        data = {}
    return cast(Dict[str, Any], data)


def _read_settings_file(settings_file: "Path") -> SettingsData:
    """
    Read the settings file, which is a json object with settings IDs as keys
    and boolean values. For each key, look up the `Settings` object with that
    key. If the key is one of the old IDs (kebab case), replace it with the
    new ID and rewrite the settings file

    :param settings_file: the path to the settings file
    :return: a dict with all new settings IDs as the keys, and boolean values
        (the values stored in the settings file, or `False` if the key was not
        found). Along with the version.
    """
    # Read settings from persistent file
    data = _read_json_file(settings_file)
    settings, version = _migrate(data)
    settings = _ensure(settings)

    if data.get("_version") != version:
        _write_settings_file(settings, version, settings_file)

    return SettingsData(settings_map=settings, version=version)


def _write_settings_file(
    data: Mapping[str, Any],
    version: int,
    settings_file: "Path",
) -> None:
    try:
        with settings_file.open("w") as fd:
            json.dump({**data, "_version": version}, fd)
            fd.flush()
            os.fsync(fd.fileno())
    except OSError:
        log.exception(f"Failed to write advanced settings file to {settings_file}")


def _migrate0to1(previous: Mapping[str, Any]) -> SettingsMap:
    """
    Migrate to version 1 of the feature flags file. Replaces old IDs with new
    IDs and sets any False values to None
    """
    next: SettingsMap = {}

    for s in settings:
        id = s.id
        old_id = s.old_id

        if previous.get(id) is True:
            next[id] = True
        elif old_id and previous.get(old_id) is True:
            next[id] = True
        else:
            next[id] = None

    return next


def _migrate1to2(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 2 of the feature flags file. Adds the
    disableLogAggregation config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["disableLogAggregation"] = None
    return newmap


def _migrate2to3(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 3 of the feature flags file. Adds the
    enableApi1BackCompat config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableApi1BackCompat"] = None
    newmap["useProtocolApi2"] = None
    return newmap


def _migrate3to4(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 4 of the feature flags file. Adds the
    useV1HttpApi config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["useV1HttpApi"] = None
    return newmap


def _migrate4to5(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 5 of the feature flags file. Adds the
    enableDoorSafetyFeature config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableDoorSafetySwitch"] = None
    return newmap


def _migrate5to6(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 6 of the feature flags file. Adds the
    enableTipLengthCalibration config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableTipLengthCalibration"] = None
    return newmap


def _migrate6to7(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 7 of the feature flags file. Adds the
    enableHttpProtocolSessions config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableHttpProtocolSessions"] = None
    return newmap


def _migrate7to8(previous: SettingsMap) -> SettingsMap:
    """
    Migration to version 8 of the feature flags file. Adds the
    enableFastProtocolUpload config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableFastProtocolUpload"] = None
    return newmap


def _migrate8to9(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 9 of the feature flags file.

    - Adds the enableProtocolEngine config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableProtocolEngine"] = None
    return newmap


def _migrate9to10(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 10 of the feature flags file.

    - Removes deprecated useProtocolApi2 option
    - Removes deprecated enableApi1BackCompat option
    - Removed deprecated useV1HttpApi option
    - Removes deprecated enableTipLengthCalibration option
    - Removes deprecated enableFastProtocolUpload option
    - Adds disableFastProtocolUpload option
    """
    removals = [
        "useProtocolApi2",
        "enableApi1BackCompat",
        "useV1HttpApi",
        "enableTipLengthCalibration",
        "enableFastProtocolUpload",
    ]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    newmap["disableFastProtocolUpload"] = None
    return newmap


def _migrate10to11(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 11 of the feature flags file.

    - Removes deprecated enableProtocolEngine option
    """
    removals = ["enableProtocolEngine"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate11to12(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 12 of the feature flags file.

    - Adds the enableOT3HardwareController config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableOT3HardwareController"] = None
    return newmap


def _migrate12to13(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 13 of the feature flags file.

    - Removes deprecated calibrateToBottom option
    """
    removals = ["calibrateToBottom"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate13to14(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 14 of the feature flags file.

    - Removes deprecated enableHttpProtocolSessions option
    """
    removals = ["enableHttpProtocolSessions"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate14to15(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 15 of the feature flags file.

    - adds enableHeaterShakerPAPI option
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableHeaterShakerPAPI"] = None
    return newmap


def _migrate15to16(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 16 of the feature flags file.

    - Removes deprecated enableHeaterShakerPAPI option
    """
    removals = ["enableHeaterShakerPAPI"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate16to17(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 17 of the advanced settings file.

    - Adds enableProtocolEnginePAPICore option
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableProtocolEnginePAPICore"] = None
    return newmap


def _migrate17to18(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 18 of the advanced settings file.

    - Adds enableLoadLiquid option
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableLoadLiquid"] = None
    return newmap


def _migrate18to19(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 19 of the feature flags file.

    - Removes deprecated enableLoadLiquid option
    """
    removals = ["enableLoadLiquid"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate19to20(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 20 of the feature flags file.

    - Adds the enableOT3FirmwareUpdates config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableOT3FirmwareUpdates"] = None
    return newmap


def _migrate20to21(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 21 of the feature flags file.

    - Removes deprecated enableProtocolEnginePAPICore option
    """
    removals = ["enableProtocolEnginePAPICore"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate21to22(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 22 of the feature flags file.

    - Removes deprecated enableOT3FirmwareUpdates option
    """
    removals = ["enableOT3FirmwareUpdates"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate22to23(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 23 of the feature flags file.

    - Adds the rearPanelIntegration config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["rearPanelIntegration"] = None
    return newmap


def _migrate23to24(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 24 of the feature flags file.

    - flips the rearPanelIntegration config element default to true.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["rearPanelIntegration"] = True
    return newmap


def _migrate24to25(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 25 of the feature flags file.

    - Adds the disableStallDetection config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["disableStallDetection"] = None
    return newmap


def _migrate25to26(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 26 of the feature flags file.

    - Adds the disableStatusBar config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["disableStatusBar"] = None
    return newmap


def _migrate26to27(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 27 of the feature flags file.

    - Adds the disableOverpressureDetection config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["disableOverpressureDetection"] = None
    return newmap


def _migrate27to28(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 28 of the feature flags file.

    - Adds the disableTipPresenceDetection config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["disableTipPresenceDetection"] = None
    return newmap


def _migrate28to29(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 29 of the feature flags file.

    - Adds the estopNotRequired config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["estopNotRequired"] = None
    return newmap


def _migrate29to30(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 30 of the feature flags file.

    - Removes the disableTipPresenceDetection flag.
    """
    return {k: v for k, v in previous.items() if "disableTipPresenceDetection" != k}


def _migrate30to31(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 31 of the feature flags file.

    - Adds the enableErrorRecoveryExperiments config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableErrorRecoveryExperiments"] = None
    return newmap


def _migrate31to32(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 32 of the feature flags file.

    - Adds the enableOEMMode config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enableOEMMode"] = None
    return newmap


def _migrate32to33(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 33 of the feature flags file.

    - Adds the enablePerformanceMetrics config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["enablePerformanceMetrics"] = None
    return newmap


def _migrate33to34(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 34 of the feature flags file.

    - Removes disableFastProtocolUpload
    """
    removals = ["disableFastProtocolUpload"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate34to35(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 35 of the feature flags file.

    - Removes disableLogAggregation
    """
    removals = ["disableLogAggregation"]
    newmap = {k: v for k, v in previous.items() if k not in removals}
    return newmap


def _migrate35to36(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 36 of the feature flags file.

    - Adds the allowLiquidClasses config element.
    """
    newmap = {k: v for k, v in previous.items()}
    newmap["allowLiquidClasses"] = None
    return newmap


def _migrate36to37(previous: SettingsMap) -> SettingsMap:
    """Migrate to version 37 of the feature flags file.

    - Removes the allowLiquidClasses flag.
    """
    return {k: v for k, v in previous.items() if "allowLiquidClasses" != k}


_MIGRATIONS = [
    _migrate0to1,
    _migrate1to2,
    _migrate2to3,
    _migrate3to4,
    _migrate4to5,
    _migrate5to6,
    _migrate6to7,
    _migrate7to8,
    _migrate8to9,
    _migrate9to10,
    _migrate10to11,
    _migrate11to12,
    _migrate12to13,
    _migrate13to14,
    _migrate14to15,
    _migrate15to16,
    _migrate16to17,
    _migrate17to18,
    _migrate18to19,
    _migrate19to20,
    _migrate20to21,
    _migrate21to22,
    _migrate22to23,
    _migrate23to24,
    _migrate24to25,
    _migrate25to26,
    _migrate26to27,
    _migrate27to28,
    _migrate28to29,
    _migrate29to30,
    _migrate30to31,
    _migrate31to32,
    _migrate32to33,
    _migrate33to34,
    _migrate34to35,
    _migrate35to36,
    _migrate36to37,
]
"""
List of all migrations to apply, indexed by (version - 1). See _migrate below
for how the migration functions are applied. Each migration function should
return a new dictionary (rather than modify their input)
"""


def _migrate(data: Mapping[str, Any]) -> SettingsData:
    """
    Check the version integer of the JSON file data a run any necessary
    migrations to get us to the latest file format. Returns dictionary of
    settings and version migrated to
    """
    next = dict(data)
    version = next.pop("_version", 0)
    target_version = len(_MIGRATIONS)
    migrations = _MIGRATIONS[version:]

    if len(migrations) > 0:
        log.info(
            "Migrating advanced settings from version {} to {}".format(
                version, target_version
            )
        )

    for m in migrations:
        next = m(next)

    return SettingsData(settings_map=next, version=target_version)


def _ensure(data: Mapping[str, Any]) -> SettingsMap:
    """Ensure config data is valid, regardless of version.

    Even after migration, we may have an invalid file. For instance,
    the user may have _downgraded_. Make sure all required keys are present,
    regardless of config version.
    """
    newdata = {k: v for k, v in data.items()}
    for s in settings:
        if s.id not in newdata:
            newdata[s.id] = None
    return newdata


def get_setting_with_env_overload(setting_name: str, robot_type: RobotTypeEnum) -> bool:
    env_name = "OT_API_FF_" + setting_name
    defn = get_setting_definition(setting_name)
    if env_name in os.environ:
        return os.environ[env_name].lower() in {"1", "true", "on"}
    else:
        s = get_adv_setting(setting_name, robot_type)
        if s is not None:
            return s.value is True
        if defn is None:
            return False
        if robot_type in defn.default_true_on_robot_types:
            return True
        return False


_SETTINGS_RESTART_REQUIRED = False
# This is a bit of global state that indicates whether a setting has changed
# that requires a restart. It's OK for this to be global because the behavior
# it's catching is global - changing the kind of setting that requires a
# a restart anywhere, even if you theoretically have two servers running in
# the same process, will require _all_ of them to be restarted.


def is_restart_required() -> bool:
    return _SETTINGS_RESTART_REQUIRED


def set_restart_required() -> bool:
    """Set the restart required flag."""
    global _SETTINGS_RESTART_REQUIRED
    _SETTINGS_RESTART_REQUIRED = True
    return _SETTINGS_RESTART_REQUIRED



"""Legacy Protocol API module implementation logic."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, List, Optional, cast

from opentrons.drivers.types import (
    HeaterShakerLabwareLatchStatus,
    ThermocyclerLidStatus,
)
from opentrons.hardware_control import SynchronousAdapter, modules as hw_modules
from opentrons.hardware_control.types import Axis
from opentrons.hardware_control.modules.types import (
    ModuleModel,
    TemperatureStatus,
    MagneticStatus,
    SpeedStatus,
    MagneticModuleModel,
    ThermocyclerStep,
)
from opentrons.types import DeckSlotName, Location


from ..module import (
    AbstractModuleCore,
    AbstractTemperatureModuleCore,
    AbstractMagneticModuleCore,
    AbstractThermocyclerCore,
    AbstractHeaterShakerCore,
)

from .legacy_labware_core import LegacyLabwareCore
from .module_geometry import ModuleGeometry, ThermocyclerGeometry, HeaterShakerGeometry
from ...labware import Labware

if TYPE_CHECKING:
    from .legacy_protocol_core import LegacyProtocolCore


_log = logging.getLogger(__name__)


class NoTargetTemperatureSetError(RuntimeError):
    """An error raised when awaiting temperature when no target was set."""


class CannotPerformModuleAction(RuntimeError):
    """An error raised when attempting to execute an invalid module action."""


class LegacyModuleCore(AbstractModuleCore):
    """Legacy ModuleCore implementation for pre-ProtocolEngine protocols."""

    def __init__(
        self,
        sync_module_hardware: SynchronousAdapter[hw_modules.AbstractModule],
        requested_model: ModuleModel,
        geometry: ModuleGeometry,
        protocol_core: LegacyProtocolCore,
    ) -> None:
        self._sync_module_hardware = sync_module_hardware
        self._requested_model = requested_model
        self._geometry = geometry
        self._protocol_core = protocol_core

    @property
    def geometry(self) -> ModuleGeometry:
        return self._geometry

    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return self._geometry.model

    def get_requested_model(self) -> ModuleModel:
        """Get the model identifier the module was requested as.

        This may differ from the actual model returned by `get_model`.
        """
        return self._requested_model

    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        device_info = self._sync_module_hardware.device_info
        return cast(str, device_info["serial"])

    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""
        return DeckSlotName.from_primitive(self._geometry.parent)  # type: ignore[arg-type]

    def get_deck_slot_id(self) -> str:
        return self.get_deck_slot().id

    def get_display_name(self) -> str:
        """Get the module's display name."""
        return self._geometry.display_name

    def add_labware_core(self, labware_core: LegacyLabwareCore) -> Labware:
        """Add a labware to the module."""
        labware = self.geometry.add_labware(
            Labware(
                core=labware_core,
                api_version=self._protocol_core.api_version,
                protocol_core=None,  # type: ignore[arg-type]
                core_map=None,  # type: ignore[arg-type]
            )
        )
        self._protocol_core.get_deck().recalculate_high_z()
        return labware


class LegacyTemperatureModuleCore(LegacyModuleCore, AbstractTemperatureModuleCore):
    """Legacy core control implementation for an attached Temperature Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.TempDeck]

    def set_target_temperature(self, celsius: float) -> None:
        """Set the Temperature Module's target temperature in °C."""
        self._sync_module_hardware.start_set_temperature(celsius)

    def wait_for_target_temperature(self, celsius: Optional[float] = None) -> None:
        """Wait until the module's target temperature is reached.

        Specifying a value for ``celsius`` that is different than
        the module's current target temperature may behave unpredictably.
        """
        self._sync_module_hardware.await_temperature(celsius)

    def deactivate(self) -> None:
        """Deactivate the Temperature Module."""
        self._sync_module_hardware.deactivate()

    def get_current_temperature(self) -> float:
        """Get the module's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the module's target temperature in °C, if set."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_status(self) -> TemperatureStatus:
        """Get the module's current temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]


class LegacyMagneticModuleCore(LegacyModuleCore, AbstractMagneticModuleCore):
    """Core control interface for an attached Magnetic Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.MagDeck]

    def engage(
        self,
        height_from_base: Optional[float] = None,
        height_from_home: Optional[float] = None,
    ) -> None:
        """Raise the module's magnets.

        Only one of `height_from_base` or `height_from_home` may be specified.

        Args:
            height_from_base: Distance from labware base to raise the magnets.
            height_from_home: Distance from motor home position to raise the magnets.
        """
        # TODO(mc, 2022-09-23): update when HW API uses real millimeters
        # https://opentrons.atlassian.net/browse/RET-1242
        if height_from_base is not None:
            self._sync_module_hardware.engage(height_from_base=height_from_base)
        else:
            assert height_from_home is not None, "Engage height must be specified"
            self._sync_module_hardware.engage(height=height_from_home)

    def engage_to_labware(
        self,
        offset: float = 0,
        preserve_half_mm: bool = False,
    ) -> None:
        """Raise the module's magnets up to its loaded labware.

        Args:
            offset: Offset from the labware's default engage height.
            preserve_half_mm: Preserve any values that may accidentally be in half-mm,
                passing them directly onwards to the hardware control API.

        Raises:
            ValueError: Labware is not loaded or has no default engage height.
        """
        labware = self._geometry.labware

        if labware is None:
            raise ValueError(
                "No labware loaded in Magnetic Module;"
                " you must specify an engage height explicitly"
                " using `height_from_base` or `height`"
            )

        labware_core = cast(LegacyLabwareCore, labware._core)
        engage_height = labware_core.get_default_magnet_engage_height(preserve_half_mm)

        if engage_height is None:
            raise ValueError(
                f"Currently loaded labware {labware} does not have"
                " a default engage height; specify engage height explicitly"
                " using `height_from_base` or `height`"
            )

        if (
            self._geometry.model == MagneticModuleModel.MAGNETIC_V1
            and not preserve_half_mm
        ):
            engage_height *= 2

        # TODO(mc, 2022-09-23): use real millimeters instead of model-dependent mm
        # https://opentrons.atlassian.net/browse/RET-1242
        # TODO(mc, 2022-09-23): use `height_from_base` to match JSONv5
        # https://opentrons.atlassian.net/browse/RSS-110
        self._sync_module_hardware.engage(height=engage_height + offset)

    def disengage(self) -> None:
        """Lower the magnets back into the module."""
        self._sync_module_hardware.deactivate()

    def get_status(self) -> MagneticStatus:
        """Get the module's current magnet status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]

    def add_labware_core(self, labware_core: LegacyLabwareCore) -> Labware:
        """Add a labware to the module."""
        labware = super().add_labware_core(labware_core)
        if labware_core.get_default_magnet_engage_height() is None:
            name = labware_core.get_name()
            _log.warning(
                f"The labware definition for {name} does not define a"
                " default engagement height for use with the Magnetic Module;"
                " you must specify a height explicitly when calling engage()."
            )
        return labware


class LegacyThermocyclerCore(LegacyModuleCore, AbstractThermocyclerCore):
    """Core control interface for an attached Thermocycler Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.Thermocycler]
    _geometry: ThermocyclerGeometry

    def open_lid(self) -> ThermocyclerLidStatus:
        """Open the thermocycler's lid."""
        self._prepare_for_lid_move()
        self._geometry.lid_status = self._sync_module_hardware.open()
        return self._geometry.lid_status

    def close_lid(self) -> ThermocyclerLidStatus:
        """Close the thermocycler's lid."""
        self._prepare_for_lid_move()
        self._geometry.lid_status = self._sync_module_hardware.close()
        return self._geometry.lid_status

    def set_target_block_temperature(
        self,
        celsius: float,
        hold_time_seconds: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Set the target temperature for the well block, in °C."""
        self._sync_module_hardware.set_target_block_temperature(
            celsius=celsius,
            hold_time_seconds=hold_time_seconds,
            volume=block_max_volume,
        )

    def wait_for_block_temperature(self) -> None:
        """Wait for target block temperature to be reached."""
        self._sync_module_hardware.wait_for_block_target()

    def set_target_lid_temperature(self, celsius: float) -> None:
        """Set the target temperature for the heated lid, in °C."""
        self._sync_module_hardware.set_target_lid_temperature(celsius=celsius)

    def wait_for_lid_temperature(self) -> None:
        """Wait for target lid temperature to be reached."""
        self._sync_module_hardware.wait_for_lid_target()

    def execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Execute a Thermocycler Profile."""
        self._sync_module_hardware.cycle_temperatures(
            steps=steps, repetitions=repetitions, volume=block_max_volume
        )

    def deactivate_lid(self) -> None:
        """Turn off the heated lid."""
        self._sync_module_hardware.deactivate_lid()

    def deactivate_block(self) -> None:
        """Turn off the well block temperature controller"""
        self._sync_module_hardware.deactivate_block()

    def deactivate(self) -> None:
        """Turn off the well block temperature controller, and heated lid"""
        self._sync_module_hardware.deactivate()

    def get_lid_position(self) -> Optional[ThermocyclerLidStatus]:
        """Get the thermocycler's lid position."""
        return self._sync_module_hardware.lid_status  # type: ignore[no-any-return]

    def get_block_temperature_status(self) -> TemperatureStatus:
        """Get the thermocycler's block temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]

    def get_lid_temperature_status(self) -> Optional[TemperatureStatus]:
        """Get the thermocycler's lid temperature status."""
        return self._sync_module_hardware.lid_temp_status  # type: ignore[no-any-return]

    def get_block_temperature(self) -> Optional[float]:
        """Get the thermocycler's current block temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_block_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target block temperature in °C."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_lid_temperature(self) -> Optional[float]:
        """Get the thermocycler's current lid temperature in °C."""
        return self._sync_module_hardware.lid_temp  # type: ignore[no-any-return]

    def get_lid_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target lid temperature in °C."""
        return self._sync_module_hardware.lid_target  # type: ignore[no-any-return]

    def get_ramp_rate(self) -> Optional[float]:
        """Get the thermocycler's current ramp rate in °C/sec."""
        return self._sync_module_hardware.ramp_rate  # type: ignore[no-any-return]

    def get_hold_time(self) -> Optional[float]:
        """Get the remaining hold time in seconds."""
        return self._sync_module_hardware.hold_time  # type: ignore[no-any-return]

    def get_total_cycle_count(self) -> Optional[int]:
        """Get number of repetitions for current set cycle."""
        return self._sync_module_hardware.total_cycle_count  # type: ignore[no-any-return]

    def get_current_cycle_index(self) -> Optional[int]:
        """Get index of the current set cycle repetition."""
        return self._sync_module_hardware.current_cycle_index  # type: ignore[no-any-return]

    def get_total_step_count(self) -> Optional[int]:
        """Get number of steps within the current cycle."""
        return self._sync_module_hardware.total_step_count  # type: ignore[no-any-return]

    def get_current_step_index(self) -> Optional[int]:
        """Get the index of the current step within the current cycle."""
        return self._sync_module_hardware.current_step_index  # type: ignore[no-any-return]

    def flag_unsafe_move(self, to_loc: Location, from_loc: Location) -> None:
        """Check if a movement may interfere with this Thermocycler.

        Args:
            to_loc: Movement destination.
            from_loc: Movement origin.

        Raises:
            RuntimeError: The movement is unsafe.
        """
        self._geometry.flag_unsafe_move(to_loc, from_loc, self.get_lid_position())

    def _prepare_for_lid_move(self) -> None:
        loaded_instruments = [
            instr
            for mount, instr in self._protocol_core.get_loaded_instruments().items()
            if instr is not None
        ]
        try:
            instr_core = loaded_instruments[0]
        except IndexError:
            _log.warning(
                "Cannot assure a safe gantry position to avoid colliding"
                " with the lid of the Thermocycler Module."
            )
        else:
            protocol_core = self._protocol_core
            hardware = protocol_core.get_hardware()
            hardware.retract(instr_core.get_mount())
            high_point = hardware.current_position(instr_core.get_mount())
            trash_top = protocol_core.fixed_trash.get_well_core("A1").get_top(
                z_offset=0
            )
            safe_point = trash_top._replace(
                z=high_point[Axis.by_mount(instr_core.get_mount())]
            )
            instr_core.move_to(
                location=Location(safe_point, None),
                well_core=None,
                force_direct=True,
                minimum_z_height=None,
                speed=None,
            )


class LegacyHeaterShakerCore(LegacyModuleCore, AbstractHeaterShakerCore):
    """Core control interface for an attached Heater-Shaker Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.HeaterShaker]
    _geometry: HeaterShakerGeometry

    def set_target_temperature(self, celsius: float) -> None:
        """Set the labware plate's target temperature in °C."""
        self._sync_module_hardware.start_set_temperature(celsius)

    def wait_for_target_temperature(self) -> None:
        """Wait for the labware plate's target temperature to be reached."""
        target_temperature = self.get_target_temperature()

        if target_temperature is None:
            raise NoTargetTemperatureSetError(
                "Heater-Shaker Module does not have a target temperature set."
            )

        self._sync_module_hardware.await_temperature(target_temperature)

    def set_and_wait_for_shake_speed(self, rpm: int) -> None:
        """Set the shaker's target shake speed and wait for it to spin up."""
        if (
            self.get_labware_latch_status()
            == HeaterShakerLabwareLatchStatus.IDLE_CLOSED
        ):
            self._prepare_for_shake()
            self._sync_module_hardware.set_speed(rpm=rpm)
        else:
            # TODO: Figure out whether to issue close latch behind the scenes instead
            raise CannotPerformModuleAction(
                "Cannot start shaking unless labware latch is closed."
            )

    def open_labware_latch(self) -> None:
        """Open the labware latch."""
        if self.get_speed_status() != SpeedStatus.IDLE:
            # TODO: What to do when speed status is ERROR?
            raise CannotPerformModuleAction(
                """Cannot open labware latch while module is shaking."""
            )
        self._prepare_for_latch_open()
        self._sync_module_hardware.open_labware_latch()

    def close_labware_latch(self) -> None:
        """Close the labware latch."""
        self._sync_module_hardware.close_labware_latch()

    def deactivate_shaker(self) -> None:
        """Stop shaking."""
        self._sync_module_hardware.deactivate_shaker()

    def deactivate_heater(self) -> None:
        """Stop heating."""
        self._sync_module_hardware.deactivate_heater()

    def get_current_temperature(self) -> float:
        """Get the labware plate's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the labware plate's target temperature in °C, if set."""
        return self._sync_module_hardware.target_temperature  # type: ignore[no-any-return]

    def get_current_speed(self) -> int:
        """Get the shaker's current speed in RPM."""
        return self._sync_module_hardware.speed  # type: ignore[no-any-return]

    def get_target_speed(self) -> Optional[int]:
        """Get the shaker's target speed in RPM, if set."""
        return self._sync_module_hardware.target_speed  # type: ignore[no-any-return]

    def get_temperature_status(self) -> TemperatureStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.temperature_status  # type: ignore[no-any-return]

    def get_speed_status(self) -> SpeedStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.speed_status  # type: ignore[no-any-return]

    def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Get the module's labware latch status."""
        return self._sync_module_hardware.labware_latch_status  # type: ignore[no-any-return]

    def flag_unsafe_move(self, to_loc: Location, is_multichannel: bool) -> None:
        """Check if a movement may interfere with this Heater-Shaker.

        Args:
            to_loc: Movement destination.
            is_multichannel: If the pipette is a multi-channel pipette,
                which has more movement restrictions than a single-channel.

        Raises:
            TypeError: `to_loc` is an invalid destination.
            RuntimeError: The movement is unsafe.
        """
        destination_slot = to_loc.labware.first_parent()
        if destination_slot is None:
            _log.warning(
                "Pipette movement destination has no slot associated with it. Cannot"
                " determine whether movement will safely avoid colliding with the Heater-Shaker."
            )
            return

        is_labware_latch_closed = (
            self.get_labware_latch_status()
            == HeaterShakerLabwareLatchStatus.IDLE_CLOSED
        )
        is_plate_shaking = self.get_speed_status() != SpeedStatus.IDLE

        to_labware_like = to_loc.labware
        is_tiprack: bool
        if (
            to_labware_like.is_labware
        ):  # Do we consider this a valid location for move_to?
            is_tiprack = to_labware_like.as_labware().is_tiprack
        elif to_labware_like.parent.is_labware:
            is_tiprack = to_labware_like.parent.as_labware().is_tiprack
        else:
            raise TypeError(
                "Invalid destination location type. "
                "Cannot determine pipette movement safety."
            )

        self._geometry.flag_unsafe_move(
            to_slot=int(destination_slot),
            is_tiprack=is_tiprack,
            is_using_multichannel=is_multichannel,
            is_plate_shaking=is_plate_shaking,
            is_labware_latch_closed=is_labware_latch_closed,
        )

    def _prepare_for_shake(self) -> None:
        """
        Before shaking, retracts pipettes if they're parked over a slot
        adjacent to the heater-shaker.
        """
        protocol_core = self._protocol_core
        if self._geometry.is_pipette_blocking_shake_movement(
            pipette_location=protocol_core.get_last_location()
        ):
            hardware = protocol_core.get_hardware()
            hardware.home(axes=[axis for axis in Axis.ot2_mount_axes()])
            protocol_core.set_last_location(None)

    def _prepare_for_latch_open(self) -> None:
        """
        Before opening latch, retracts pipettes if they're parked over a slot
        east/ west of the heater-shaker.
        """
        protocol_core = self._protocol_core
        if self._geometry.is_pipette_blocking_latch_movement(
            pipette_location=protocol_core.get_last_location()
        ):
            hardware = protocol_core.get_hardware()
            hardware.home(axes=[axis for axis in Axis.ot2_mount_axes()])
            protocol_core.set_last_location(None)


def create_module_core(
    module_hardware_api: hw_modules.AbstractModule,
    requested_model: ModuleModel,
    geometry: ModuleGeometry,
    protocol_core: LegacyProtocolCore,
) -> LegacyModuleCore:
    core_cls = LegacyModuleCore

    if isinstance(module_hardware_api, hw_modules.TempDeck):
        core_cls = LegacyTemperatureModuleCore
    elif isinstance(module_hardware_api, hw_modules.MagDeck):
        core_cls = LegacyMagneticModuleCore
    elif isinstance(module_hardware_api, hw_modules.Thermocycler):
        core_cls = LegacyThermocyclerCore
    elif isinstance(module_hardware_api, hw_modules.HeaterShaker):
        core_cls = LegacyHeaterShakerCore

    return core_cls(
        sync_module_hardware=SynchronousAdapter(module_hardware_api),
        requested_model=requested_model,
        geometry=geometry,
        protocol_core=protocol_core,
    )



"""Script for starting up a python module emulator."""
import logging
import asyncio
from argparse import ArgumentParser
from typing import Dict, Callable
from typing_extensions import Final

from opentrons.hardware_control.emulation.abstract_emulator import AbstractEmulator
from opentrons.hardware_control.emulation.heater_shaker import HeaterShakerEmulator
from opentrons.hardware_control.emulation.types import ModuleType
from opentrons.hardware_control.emulation.magdeck import MagDeckEmulator
from opentrons.hardware_control.emulation.parser import Parser
from opentrons.hardware_control.emulation.tempdeck import TempDeckEmulator
from opentrons.hardware_control.emulation.thermocycler import ThermocyclerEmulator


from opentrons.hardware_control.emulation.run_emulator import run_emulator_client
from opentrons.hardware_control.emulation.settings import Settings, ProxySettings

emulator_builder: Final[Dict[str, Callable[[Settings], AbstractEmulator]]] = {
    ModuleType.Magnetic.value: lambda s: MagDeckEmulator(Parser(), s.magdeck),
    ModuleType.Temperature.value: lambda s: TempDeckEmulator(Parser(), s.tempdeck),
    ModuleType.Thermocycler.value: lambda s: ThermocyclerEmulator(
        Parser(), s.thermocycler
    ),
    ModuleType.Heatershaker.value: lambda s: HeaterShakerEmulator(
        Parser(), s.heatershaker
    ),
}

emulator_port: Final[Dict[str, Callable[[Settings], ProxySettings]]] = {
    ModuleType.Magnetic.value: lambda s: s.magdeck_proxy,
    ModuleType.Temperature.value: lambda s: s.temperature_proxy,
    ModuleType.Thermocycler.value: lambda s: s.thermocycler_proxy,
    ModuleType.Heatershaker.value: lambda s: s.heatershaker_proxy,
}


async def run(settings: Settings, emulator_name: str, host: str) -> None:
    """Run an emulator.

    Args:
        settings: emulator settings
        emulator_name: Name of emulator. This must be a key in emulator_builder
        host: host to connect to.

    Returns:
        None
    """
    e = emulator_builder[emulator_name](settings)
    proxy_settings = emulator_port[emulator_name](settings)
    await run_emulator_client(host, proxy_settings.emulator_port, e)


def main() -> None:
    """Entry point."""
    a = ArgumentParser()
    a.add_argument(
        "emulator",
        type=str,
        choices=emulator_builder.keys(),
        help="which module to emulate.",
    )
    a.add_argument("host", type=str, help="the emulator host")
    args = a.parse_args()

    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)
    asyncio.run(run(Settings(), args.emulator, args.host))


if __name__ == "__main__":
    main()



"""Command models for moving any robot mount to a destination point."""
from __future__ import annotations
from typing import Literal, Type, Optional, TYPE_CHECKING, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.types import MountType

from ..movement_common import DestinationPositionResult
from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from opentrons.protocol_engine.types import DeckPoint
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence


if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import MovementHandler


MoveToCommandType = Literal["robot/moveTo"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class MoveToParams(BaseModel):
    """Payload required to move to a destination position."""

    mount: MountType = Field(
        ...,
        description="The mount to move to the destination point.",
    )
    destination: DeckPoint = Field(
        ...,
        description="X, Y and Z coordinates in mm from deck's origin location (left-front-bottom corner of work space)",
    )
    speed: float | SkipJsonSchema[None] = Field(
        default=None,
        description="The max velocity to move the axes at. Will fall to hardware defaults if none provided.",
        json_schema_extra=_remove_default,
    )


class MoveToResult(DestinationPositionResult):
    """Result data from the execution of a MoveTo command."""

    pass


class MoveToImplementation(
    AbstractCommandImpl[MoveToParams, SuccessData[MoveToResult]]
):
    """MoveTo command implementation."""

    def __init__(
        self,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._movement = movement

    async def execute(self, params: MoveToParams) -> SuccessData[MoveToResult]:
        """Move to a given destination on a flex."""
        x, y, z = await self._movement.move_mount_to(
            mount=params.mount, destination=params.destination, speed=params.speed
        )
        return SuccessData(
            public=MoveToResult(position=DeckPoint(x=x, y=y, z=z)),
        )


class MoveTo(BaseCommand[MoveToParams, MoveToResult, ErrorOccurrence]):
    """MoveTo command model."""

    commandType: MoveToCommandType = "robot/moveTo"
    params: MoveToParams
    result: Optional[MoveToResult] = None

    _ImplementationCls: Type[MoveToImplementation] = MoveToImplementation


class MoveToCreate(BaseCommandCreate[MoveToParams]):
    """MoveTo command request model."""

    commandType: MoveToCommandType = "robot/moveTo"
    params: MoveToParams

    _CommandCls: Type[MoveTo] = MoveTo



"""Protocol specifying API gripper control."""
from typing import Optional
from typing_extensions import Protocol

from opentrons.hardware_control.dev_types import GripperDict
from opentrons.hardware_control.instruments.ot3.gripper import Gripper


class GripperController(Protocol):
    """A protocol specifying gripper API functions."""

    async def grip(
        self, force_newtons: Optional[float] = None, stay_engaged: bool = True
    ) -> None:
        ...

    async def home_gripper_jaw(self) -> None:
        ...

    async def ungrip(self, force_newtons: Optional[float] = None) -> None:
        """Release gripped object.

        To simply open the jaw, use `home_gripper_jaw` instead.
        """
        ...

    async def idle_gripper(self) -> None:
        """Move gripper to its idle, gripped position."""
        ...

    def gripper_jaw_can_home(self) -> bool:
        """Check if it is valid to home the gripper jaw.

        This should return False if the API believes that the gripper is
        currently holding something.
        """
        ...

    def raise_error_if_gripper_pickup_failed(
        self,
        expected_grip_width: float,
        grip_width_uncertainty_wider: float,
        grip_width_uncertainty_narrower: float,
    ) -> None:
        """Ensure that a gripper pickup succeeded."""

    @property
    def attached_gripper(self) -> Optional[GripperDict]:
        """Get a dict of all attached grippers."""
        ...

    @property
    def hardware_gripper(self) -> Optional[Gripper]:
        """Get attached gripper, if present."""
        ...



"""Unique ID generation provider."""
from datetime import datetime, timezone
from uuid import uuid4
from typing import Optional


class ModelUtils:
    """Common resource model utilities provider."""

    @staticmethod
    def generate_id(prefix: str = "") -> str:
        """Generate a unique identifier.

        Uses UUIDv4 for safety in a multiprocessing environment.

        Params:
            prefix: Prepended to the UUIDv4.
        """
        return prefix + str(uuid4())

    @staticmethod
    def ensure_id(maybe_id: Optional[str] = None) -> str:
        """Generate a unique identifier only if the given input is None.

        Uses UUIDv4 for safety in a multiprocessing environment.
        """
        return maybe_id if maybe_id is not None else ModelUtils.generate_id()

    @staticmethod
    def get_timestamp() -> datetime:
        """Get a timestamp of the current time."""
        return datetime.now(tz=timezone.utc)
